import logging
import struct
import zlib

import numpy as np
from genomic_regions import GenomicRegion

from ..regions import Genome
from ..hic import Hic
from ..pairs import ReadPairs
from ..matrix import RegionMatrixContainer, Edge
from ..config import config
from ..tools.files import tmp_file_name
from ..tools.general import str_to_int

import os
import subprocess
import warnings
import tempfile
import gzip
import shutil

from collections import defaultdict

logger = logging.getLogger(__name__)


def is_juicer(file_name):
    try:
        if "@" in file_name:
            hic_file, _ = file_name.split("@")
            file_name = hic_file
        with open(file_name, 'rb') as req:
            try:
                magic_string = struct.unpack('<3s', req.read(3))[0]
            except struct.error:
                return False

            if magic_string != b"HIC":
                return False
            else:
                return True
    except KeyError:
        return False


def convert_juicer_to_hic(juicer_file, genome_file, resolution, juicer_tools_jar_path=None,
                          norm='NONE', output_file=None, inter_chromosomal=True,
                          chromosomes=None):
    if juicer_tools_jar_path is None:
        juicer_tools_jar_path = config.juicer_tools_jar_path
    if juicer_tools_jar_path is None:
        raise ValueError("Just must provide the juicer tools jar path or set it in fanc.conf!")

    hic = Hic(file_name=output_file, mode='w')

    # regions
    logger.info("Building genome {}".format(genome_file))
    genome = Genome.from_string(genome_file)
    regions = genome.get_regions(resolution, chromosomes=chromosomes)
    hic.add_regions(regions)
    genome.close()
    regions.close()

    if chromosomes is None:
        chromosomes = hic.chromosomes()
    chromosome_bins = hic.chromosome_bins

    logger.info("Extracting edges from hic file")
    nan_counter = 0
    for i in range(len(chromosomes)):
        chromosome1 = chromosomes[i]
        offset1 = chromosome_bins[chromosome1][0]

        for j in range(i, len(chromosomes)):
            if i != j and not inter_chromosomal:
                continue
            chromosome2 = chromosomes[j]
            offset2 = chromosome_bins[chromosome2][0]

            logger.info("{} -- {}".format(chromosome1, chromosome2))

            juicer_command = ['java', '-jar', juicer_tools_jar_path,
                              'dump', 'observed', norm, juicer_file,
                              chromosome1, chromosome2, 'BP', str(resolution)]

            juicer_process = subprocess.Popen(juicer_command, stdout=subprocess.PIPE)

            for line in juicer_process.stdout:
                fields = line.rstrip().split()

                try:
                    start, end, weight = int(fields[0]), int(fields[1]), float(fields[2])
                except ValueError:
                    continue

                start_ix = int(start / resolution) + offset1
                end_ix = int(end / resolution) + offset2
                if start_ix > end_ix:
                    start_ix, end_ix = end_ix, start_ix

                if not np.isfinite(weight):
                    nan_counter += 1
                    continue

                hic.add_edge([start_ix, end_ix, weight], check_nodes_exist=False)

    hic.flush()

    if nan_counter > 0:
        logger.warning("{} contacts could not be imported, "
                       "because they had non-finite values.".format(nan_counter))

    return hic


def to_juicer(pairs, juicer_file, juicer_tools_jar_path=None,
              resolutions=None, fragment_map=False, tmp=False,
              verbose=False):
    tmpdir = None if not tmp else tempfile.mkdtemp()
    try:
        if juicer_tools_jar_path is None:
            juicer_tools_jar_path = config.juicer_tools_jar_path
        if juicer_tools_jar_path is None:
            raise ValueError("Just must provide the juicer tools jar path or set it in fanc.conf!")

        # check that juicer tools can be executed
        if not os.path.exists(juicer_tools_jar_path):
            raise ValueError("{} does not exist or is unreadable".format(juicer_tools_jar_path))

        help_command = ['java', '-jar', juicer_tools_jar_path, '-h']
        logger.debug("Testing juicer: {}".format(" ".join(help_command)))
        res = subprocess.call(help_command)
        if res != 0:
            raise RuntimeError("juicer had nonzero exit status ({})! "
                               "It looks like your .jar path is either wrong ".format(res))

        if isinstance(pairs, ReadPairs):
            pairs = [pairs]

        chromosomes = pairs[0].chromosomes()

        with tempfile.NamedTemporaryFile(suffix='.chrom.sizes', mode='w', delete=False) as chrom_sizes_file:
            logger.info("Writing chromosome sizes")
            for chromosome, size in pairs[0].chromosome_lengths.items():
                chrom_sizes_file.write("{}\t{}\n".format(chromosome, size))
            chrom_sizes_file.flush()

            # make restriction site file
            with tempfile.NamedTemporaryFile(prefix="re_sites_", suffix='.txt', mode='w', dir=tmpdir) as re_file:
                logger.info("Making restriction map")
                for chromosome in chromosomes:
                    re_sites = []
                    for region in pairs[0].regions(chromosome, lazy=True):
                        re_sites.append(str(region.end))
                    re_file.write("{} {}\n".format(chromosome, " ".join(re_sites)))
                re_file.flush()

                logger.info("Writing pairs")
                with tempfile.NamedTemporaryFile(prefix='reads_', suffix='.txt.gz', dir=tmpdir) as f:
                    f.flush()
                    fgz = gzip.GzipFile(mode='wb', fileobj=f)

                    for chri, chromosome1 in enumerate(chromosomes):
                        for chrj in range(chri, len(chromosomes)):
                            chromosome2 = chromosomes[chrj]
                            for p in pairs:
                                for pair in p.pairs((chromosome1, chromosome2), lazy=True):
                                    if pair.left.fragment.chromosome <= pair.right.fragment.chromosome:
                                        fgz.write("{} {} {} {} {} {} {} {}\n".format(
                                            1 if pair.left.strand == -1 else 0,
                                            pair.left.fragment.chromosome,
                                            pair.left.position,
                                            pair.left.fragment.ix,
                                            1 if pair.right.strand == -1 else 0,
                                            pair.right.fragment.chromosome,
                                            pair.right.position,
                                            pair.right.fragment.ix
                                        ).encode('utf-8'))
                                    else:
                                        fgz.write("{} {} {} {} {} {} {} {}\n".format(
                                            1 if pair.right.strand == -1 else 0,
                                            pair.right.fragment.chromosome,
                                            pair.right.position,
                                            pair.right.fragment.ix,
                                            1 if pair.left.strand == -1 else 0,
                                            pair.left.fragment.chromosome,
                                            pair.left.position,
                                            pair.left.fragment.ix
                                        ).encode('utf-8'))
                    fgz.close()

                    f.flush()

                    pre_command = ['java', '-jar', juicer_tools_jar_path, 'pre']
                    if fragment_map:
                        pre_command += ['-f', re_file.name]
                    if resolutions is not None:
                        pre_command += ['-r', ",".join([str(r) for r in resolutions])]
                    if tmp:
                        pre_command += ['-t', tmpdir]
                    if verbose:
                        pre_command.append('-v')
                    pre_command += [f.name, juicer_file, chrom_sizes_file.name]

                    logger.info("Making Juicer file: {}".format(" ".join(pre_command)))
                    res = subprocess.call(pre_command)
                    if res != 0:
                        raise RuntimeError("juicer pre had nonzero exit status ({})!".format(res))
        return JuicerHic(juicer_file)
    finally:
        if tmpdir is not None:
            shutil.rmtree(tmpdir)


def _read_cstr(f):
    """
    Copyright (c) 2016 Aiden Lab

    :param f: open binary file
    :return: str
    """
    buf = ""
    while True:
        b = f.read(1)
        b = b.decode('utf-8', 'backslashreplace')
        if b is None or b == '\0':
            return str(buf)
        else:
            buf = buf + b


class LazyJuicerEdge(object):
    def __init__(self, source, sink, matrix, **kwargs):
        self._matrix = matrix
        self._weight_field = 'weight'
        self.source = source
        self.sink = sink
        self.bias = 1.
        self.expected = None

    def __getattribute__(self, item):
        if item == '_weight_field' or item != self._weight_field:
            return object.__getattribute__(self, item)

        if self.expected is None:
            return object.__getattribute__(self, item) * self.bias
        else:
            return (object.__getattribute__(self, item) * self.bias) / self.expected

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError("No such key: {}".format(item))

    @property
    def source_node(self):
        return self._matrix.region_by_ix(self.source)

    @property
    def sink_node(self):
        return self._matrix.region_by_ix(self.sink)

    @property
    def field_names(self):
        return ['weight']


class JuicerHic(RegionMatrixContainer):
    def __init__(self, hic_file, resolution=None, mode='r', tmpdir=None, norm='KR'):
        RegionMatrixContainer.__init__(self)
        if '@' in hic_file:
            hic_file, at_resolution = hic_file.split("@")
            if resolution is not None and str_to_int(at_resolution) != resolution:
                raise ValueError("Conflicting resolution specifications: "
                                 "{} and {}".format(at_resolution, resolution))
            resolution = str_to_int(at_resolution)

        if tmpdir is None or (isinstance(tmpdir, bool) and not tmpdir):
            self.tmp_file_name = None
            self._hic_file = hic_file
        else:
            logger.info("Working in temporary directory...")
            if isinstance(tmpdir, bool):
                tmpdir = tempfile.gettempdir()
            else:
                tmpdir = os.path.expanduser(tmpdir)
            self.tmp_file_name = tmp_file_name(tmpdir, prefix='tmp_fanc', extension='_juicer.hic')
            logger.info("Temporary file: {}".format(self.tmp_file_name))
            shutil.copyfile(hic_file, self.tmp_file_name)
            self._hic_file = self.tmp_file_name

        bp_resolutions, _ = self.resolutions()
        if resolution is None:
            resolution = bp_resolutions[0]
            warnings.warn("No resolution chosen for Juicer Hic - using {}bp. "
                          "Specify a custom resolution using <.hic file>@<resolution>".format(resolution))
        if resolution not in bp_resolutions:
            raise ValueError("Resolution {} not supported ({})".format(resolution, bp_resolutions))

        if mode != 'r':
            warnings.warn("Mode {} not compatible with JuicerHic. "
                          "File will be opened in read-only mode and "
                          "changes will not be saved to file!")

        self._resolution = resolution
        self._normalisation = norm
        self._unit = 'BP'
        self._mappability = None

        if not is_juicer(hic_file):
            raise ValueError("File {} does not seem to be a .hic "
                             "file produced with juicer!".format(hic_file))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return exc_type is None

    def close(self, remove_tmp=True):
        """
        Close this Juicer file and run exit operations.

        If file was opened with tmpdir in read-only mode:
        close file and delete temporary copy.

        :param remove_tmp: If False, does not delete temporary copy of file.
        """
        if self.tmp_file_name is not None and remove_tmp:
            os.remove(self.tmp_file_name)

    @property
    def version(self):
        with open(self._hic_file, 'rb') as req:
            req.read(4)  # jump to version location
            return struct.unpack('<i', req.read(4))[0]

    def _master_index(self):
        with open(self._hic_file, 'rb') as req:
            req.read(8)  # jump to master index location
            return struct.unpack('<q', req.read(8))[0]

    @staticmethod
    def _skip_to_attributes(req):
        req.seek(0)

        # skip magic, version, master
        req.read(16)

        # skip genome
        while req.read(1).decode("utf-8") != '\0':
            pass

    @staticmethod
    def _skip_to_chromosome_lengths(req):
        JuicerHic._skip_to_attributes(req)

        # skip attributes
        n_attributes = struct.unpack('<i', req.read(4))[0]
        for _ in range(0, n_attributes):
            # skip key
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass
            # skip value
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass

    @staticmethod
    def _skip_to_resolutions(req):
        JuicerHic._skip_to_chromosome_lengths(req)

        n_chromosomes = struct.unpack('<i', req.read(4))[0]
        for _ in range(0, n_chromosomes):
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass
            req.read(4)

    @staticmethod
    def _skip_to_chromosome_sites(req):
        JuicerHic._skip_to_resolutions(req)

        n_resolutions = struct.unpack('<i', req.read(4))[0]
        for _ in range(0, n_resolutions):
            req.read(4)

        n_fragment_resolutions = struct.unpack('<i', req.read(4))[0]
        for _ in range(0, n_fragment_resolutions):
            req.read(4)

    @property
    def juicer_attributes(self):
        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_attributes(req)
            attributes = {}
            n_attributes = struct.unpack('<i', req.read(4))[0]
            for _ in range(0, n_attributes):
                key = _read_cstr(req)
                value = _read_cstr(req)
                attributes[key] = value
        return attributes

    @property
    def chromosome_lengths(self):
        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_chromosome_lengths(req)

            chromosome_lengths = {}
            n_chromosomes = struct.unpack('<i', req.read(4))[0]
            for _ in range(0, n_chromosomes):
                name = _read_cstr(req)
                length = struct.unpack('<i', req.read(4))[0]
                chromosome_lengths[name] = length

        return chromosome_lengths

    def _all_chromosomes(self):
        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_chromosome_lengths(req)

            chromosomes = []
            n_chromosomes = struct.unpack('<i', req.read(4))[0]
            for _ in range(0, n_chromosomes):
                name = _read_cstr(req)
                req.read(4)
                chromosomes.append(name)

        return chromosomes

    def chromosomes(self):
        chromosomes = []
        for chromosome in self._all_chromosomes():
            if chromosome.lower() != 'all':
                chromosomes.append(chromosome)

        return chromosomes

    def resolutions(self):
        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_resolutions(req)

            resolutions = []
            n_resolutions = struct.unpack('<i', req.read(4))[0]
            for _ in range(0, n_resolutions):
                resolution = struct.unpack('<i', req.read(4))[0]
                resolutions.append(resolution)

            fragment_resolutions = []
            n_fragment_resolutions = struct.unpack('<i', req.read(4))[0]
            for _ in range(0, n_fragment_resolutions):
                resolution = struct.unpack('<i', req.read(4))[0]
                fragment_resolutions.append(resolution)

            return resolutions, fragment_resolutions

    @staticmethod
    def _skip_to_footer(req):
        req.seek(0)
        req.read(8)  # jump to master index location
        master_index = struct.unpack('<q', req.read(8))[0]

        # jump to footer location
        req.seek(master_index)
        req.read(4)  # skip number of bytes

    @staticmethod
    def _skip_to_expected_values(req):
        JuicerHic._skip_to_footer(req)

        n_entries = struct.unpack('<i', req.read(4))[0]
        for _ in range(n_entries):
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass
            req.read(12)

    @staticmethod
    def _skip_to_normalised_expected_values(req):
        JuicerHic._skip_to_expected_values(req)

        n_vectors = struct.unpack('<i', req.read(4))[0]
        for _ in range(n_vectors):
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass
            req.read(4)

            n_values = struct.unpack('<i', req.read(4))[0]
            for j in range(n_values):
                req.read(8)

            n_scaling_factors = struct.unpack('<i', req.read(4))[0]
            for _ in range(n_scaling_factors):
                req.read(12)

    @staticmethod
    def _skip_to_normalisation_vectors(req):
        JuicerHic._skip_to_normalised_expected_values(req)

        n_vectors = struct.unpack('<i', req.read(4))[0]
        for _ in range(n_vectors):
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass
            while req.read(1).decode("utf-8", "backslashreplace") != '\0':
                pass
            req.read(4)

            n_values = struct.unpack('<i', req.read(4))[0]
            for j in range(n_values):
                req.read(8)

            n_scaling_factors = struct.unpack('<i', req.read(4))[0]
            for _ in range(n_scaling_factors):
                req.read(12)

    def _matrix_positions(self):
        """
        Copyright (c) 2016 Aiden Lab
        """

        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_footer(req)

            chromosome_pair_positions = {}
            n_entries = struct.unpack('<i', req.read(4))[0]
            for _ in range(n_entries):
                key = tuple(_read_cstr(req).split('_'))
                file_position = struct.unpack('<q', req.read(8))[0]
                req.read(4)  # skip size in bytes
                chromosome_pair_positions[key] = file_position
            return chromosome_pair_positions

    @staticmethod
    def _expected_value_vectors_from_pos(req, normalisation=None, unit='BP'):
        expected_values = defaultdict(list)
        scaling_factors = defaultdict(dict)

        n_vectors = struct.unpack('<i', req.read(4))[0]
        for _ in range(n_vectors):
            if normalisation:
                entry_normalisation = _read_cstr(req)
                entry_unit = _read_cstr(req)
            else:
                entry_unit = _read_cstr(req)
                entry_normalisation = 'NONE'

            bin_size = struct.unpack('<i', req.read(4))[0]

            ev = []
            n_values = struct.unpack('<i', req.read(4))[0]

            for j in range(n_values):
                v = struct.unpack('<d', req.read(8))[0]
                ev.append(v)

            if entry_unit == unit and (normalisation is None or entry_normalisation == normalisation):
                expected_values[bin_size] = ev

            sf = dict()
            n_scaling_factors = struct.unpack('<i', req.read(4))[0]
            for _ in range(n_scaling_factors):
                chromosome_index = struct.unpack('<i', req.read(4))[0]
                f = struct.unpack('<d', req.read(8))[0]
                sf[chromosome_index] = f

            if normalisation is None or entry_normalisation == normalisation:
                scaling_factors[bin_size] = sf

        return expected_values, scaling_factors

    def expected_value_vector(self, chromosome, normalisation=None, resolution=None):
        if normalisation is None:
            normalisation = self._normalisation

        if resolution is None:
            resolution = self._resolution

        chromosome_ix = self._all_chromosomes().index(chromosome)

        if normalisation == 'NONE':
            vectors, scaling_factors = self.expected_value_vectors()
        else:
            vectors, scaling_factors = self.normalised_expected_value_vectors(normalisation)

        try:
            sf = scaling_factors[resolution][chromosome_ix]
        except KeyError:
            warnings.warn("Cannot find an expected value scaling factor "
                          "for {}, setting to 0.".format(chromosome))
            sf = 0.0
        return np.array(vectors[resolution]) / sf

    def expected_value_vectors(self):
        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_expected_values(req)

            return JuicerHic._expected_value_vectors_from_pos(req)

    def normalised_expected_value_vectors(self, normalisation=None):
        if normalisation is None:
            normalisation = self._normalisation

        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_normalised_expected_values(req)

            return JuicerHic._expected_value_vectors_from_pos(req, normalisation=normalisation)

    def expected_values(self, selected_chromosome=None, norm=True, *args, **kwargs):
        if selected_chromosome is not None:
            if norm:
                return self.expected_value_vector(selected_chromosome, self._normalisation)
            else:
                return self.expected_value_vector(selected_chromosome, 'NONE')

        intra_expected = dict()
        for chromosome in self.chromosomes():
            if norm:
                intra_expected[chromosome] = self.expected_value_vector(chromosome, self._normalisation)
            else:
                intra_expected[chromosome] = self.expected_value_vector(chromosome, 'NONE')

        return None, intra_expected, None

    def normalisation_vector(self, chromosome, normalisation=None, resolution=None, unit=None):
        if resolution is None:
            resolution = self._resolution

        if normalisation is None:
            normalisation = self._normalisation

        if normalisation == 'NONE':
            bins = int(np.ceil(self.chromosome_lengths[chromosome]/resolution))
            return [1.0] * bins

        if unit is None:
            unit = self._unit

        chromosomes = self.chromosomes()
        chromosome_index = chromosomes.index(chromosome) + 1

        with open(self._hic_file, 'rb') as req:
            JuicerHic._skip_to_normalisation_vectors(req)

            n_entries = struct.unpack('<i', req.read(4))[0]
            for _ in range(n_entries):
                entry_normalisation = _read_cstr(req)
                entry_chromosome_index = struct.unpack('<i', req.read(4))[0]
                entry_unit = _read_cstr(req)
                entry_resolution = struct.unpack('<i', req.read(4))[0]
                file_position = struct.unpack('<q', req.read(8))[0]
                req.read(4)  # skip size in bytes

                if (entry_chromosome_index == chromosome_index and
                        entry_normalisation == normalisation and
                        entry_resolution == resolution and
                        entry_unit == unit):
                    req.seek(file_position)
                    vector = []
                    n_values = struct.unpack('<i', req.read(4))[0]
                    for _ in range(n_values):
                        v = struct.unpack('<d', req.read(8))[0]
                        vector.append(v)

                    return vector
        raise ValueError("Cannot find normalisation vector that matches "
                         "chromosome: {}, normalisation: {}, "
                         "resolution: {}, unit: {}".format(chromosome, normalisation, resolution, unit))

    def region_by_ix(self, ix):
        chromosome_lengths = self.chromosome_lengths

        offset_ix = 0
        remaining_ix = ix
        chromosomes = self.chromosomes()
        current_chromosome = None
        for chromosome in chromosomes:
            current_chromosome = chromosome
            chromosome_length = chromosome_lengths[chromosome]
            if chromosome.lower() == 'all':
                continue

            ixs = int(np.ceil(chromosome_length / self._resolution))
            if remaining_ix > ixs:
                offset_ix += ixs
                remaining_ix -= ixs
            else:
                break

        region_ix = offset_ix + remaining_ix
        start = remaining_ix * self._resolution + 1
        return GenomicRegion(chromosome=current_chromosome, start=start,
                             end=min(start + self._resolution - 1,
                                     chromosome_lengths[current_chromosome]),
                             ix=region_ix)

    def _chromosome_ix_offset(self, target_chromosome):
        chromosome_lengths = self.chromosome_lengths
        if target_chromosome not in chromosome_lengths:
            raise ValueError("Chromosome {} not in matrix.".format(target_chromosome))

        offset_ix = 0
        chromosomes = self.chromosomes()
        for chromosome in chromosomes:
            chromosome_length = chromosome_lengths[chromosome]
            if chromosome.lower() == 'all':
                continue

            if target_chromosome == chromosome:
                return offset_ix

            ixs = int(np.ceil(chromosome_length / self._resolution))
            offset_ix += ixs

    def _region_start(self, region):
        region = self._convert_region(region)
        offset_ix = self._chromosome_ix_offset(region.chromosome)
        region_start = region.start if region.start is not None else 1
        ix = int((region_start - 1) / self._resolution)
        start = self._resolution * ix + 1
        return offset_ix + ix, start

    def _region_iter(self, *args, **kwargs):
        chromosome_lengths = self.chromosome_lengths

        chromosomes = self.chromosomes()
        for chromosome in chromosomes:
            chromosome_length = chromosome_lengths[chromosome]
            if chromosome.lower() == 'all':
                continue

            offset_ix = self._chromosome_ix_offset(chromosome)

            try:
                norm = self.normalisation_vector(chromosome)
            except ValueError:
                warnings.warn("Cannot find bias vector ({}, {} resolution) for chromosome {}."
                              "Continuing by masking the corresponding regions. If this is unexpected, "
                              "try choosing another normalisation method!".format(self._normalisation,
                                                                                  self._resolution, chromosome))
                norm = None

            for i, start in enumerate(range(1, chromosome_length, self._resolution)):
                if norm is None or np.isnan(norm[i]):
                    valid = False
                    bias = 1.0
                else:
                    valid = True
                    try:
                        bias = 1/norm[i]
                    except IndexError:
                        bias = 1.0

                if np.isnan(bias):
                    bias = 1.0

                end = min(start + self._resolution - 1, chromosome_length)
                region = GenomicRegion(chromosome=chromosome, start=start,
                                       end=end, bias=bias, valid=valid,
                                       ix=int(offset_ix + i))
                yield region

    def _region_subset(self, region, *args, **kwargs):
        subset_ix, subset_start = self._region_start(region)

        cl = self.chromosome_lengths[region.chromosome]
        norm = self.normalisation_vector(region.chromosome)
        for i, start in enumerate(range(subset_start, region.end, self._resolution)):
            end = min(start + self._resolution - 1, cl, region.end)
            bias_ix = int(start / self._resolution)

            if np.isnan([i]):
                valid = False
                bias = 1.0
            else:
                valid = True
                try:
                    bias = 1/norm[bias_ix]
                except IndexError:
                    bias = 1.0

            if np.isnan(bias):
                bias = 1.0

            r = GenomicRegion(chromosome=region.chromosome, start=start,
                              end=end, bias=bias, valid=valid,
                              ix=int(subset_ix + i))
            yield r

    def _region_len(self):
        length = 0
        for chromosome, chromosome_length in self.chromosome_lengths.items():
            if chromosome.lower() == 'all':
                continue

            ixs = int(np.ceil(chromosome_length / self._resolution))
            length += ixs
        return length

    def _read_block(self, req, file_position, block_size_in_bytes):
        req.seek(file_position)
        block_compressed = req.read(block_size_in_bytes)
        block = zlib.decompress(block_compressed)

        n_records = struct.unpack('<i', block[0:4])[0]
        if self.version < 7:
            for i in range(n_records):
                x = struct.unpack('<i', block[(12 * i + 4):(12 * i + 8)])[0]
                y = struct.unpack('<i', block[(12 * i + 8):(12 * i + 12)])[0]
                weight = struct.unpack('<f', block[(12 * i + 12):(12 * i + 16)])[0]
                yield x, y, weight
        else:
            x_offset = struct.unpack('<i', block[4:8])[0]
            y_offset = struct.unpack('<i', block[8:12])[0]
            use_short = not struct.unpack('<b', block[12:13])[0] == 0
            block_type = struct.unpack('<b', block[13:14])[0]
            index = 0

            if block_type == 1:
                row_count = struct.unpack('<h', block[14:16])[0]
                temp = 16
                for i in range(row_count):
                    y_raw = struct.unpack('<h', block[temp:(temp + 2)])[0]
                    temp += 2
                    y = y_raw + y_offset
                    col_count = struct.unpack('<h', block[temp:(temp + 2)])[0]
                    temp += 2
                    for j in range(col_count):
                        x_raw = struct.unpack('<h', block[temp:(temp + 2)])[0]
                        temp += 2
                        x = x_offset + x_raw
                        if not use_short:
                            weight = struct.unpack('<h', block[temp:(temp + 2)])[0]
                            temp += 2
                        else:
                            weight = struct.unpack('<f', block[temp:(temp + 4)])[0]
                            temp += 4
                        yield x, y, weight

                        index += 1
            elif block_type == 2:
                temp = 14
                n_points = struct.unpack('<i', block[temp:(temp + 4)])[0]
                temp += 4
                w = struct.unpack('<h', block[temp:(temp + 2)])[0]
                temp += 2
                for i in range(n_points):
                    row = int(i / w)
                    col = i - row * w
                    x = int(x_offset + col)
                    y = int(y_offset + row)
                    if not use_short:
                        weight = struct.unpack('<h', block[temp:(temp + 2)])[0]
                        temp += 2
                        if weight != -32768:
                            yield x, y, weight
                            index += 1
                    else:
                        weight = struct.unpack('<f', block[temp:(temp + 4)])[0]
                        temp += 4
                        if weight != 0x7fc00000:
                            yield x, y, weight
                            index = index + 1

    def _read_matrix(self, region1, region2):
        region1 = self._convert_region(region1)
        region2 = self._convert_region(region2)

        chromosomes = self._all_chromosomes()
        chromosome1_ix = chromosomes.index(region1.chromosome)
        chromosome2_ix = chromosomes.index(region2.chromosome)

        if chromosome1_ix > chromosome2_ix:
            region1, region2 = region2, region1
            chromosome1_ix, chromosome2_ix = chromosome2_ix, chromosome1_ix

        region1_chromosome_offset = self._chromosome_ix_offset(region1.chromosome)
        region2_chromosome_offset = self._chromosome_ix_offset(region2.chromosome)

        try:
            matrix_file_position = self._matrix_positions()[(str(chromosome1_ix), str(chromosome2_ix))]
        except KeyError:
            return

        with open(self._hic_file, 'rb') as req:
            req.seek(matrix_file_position)
            req.read(8)  # skip chromosome index

            block_bin_count = None
            block_column_count = None
            block_map = dict()
            n_resolutions = struct.unpack('<i', req.read(4))[0]
            for i in range(n_resolutions):
                unit = _read_cstr(req)
                req.read(20)  # skip reserved but unused fields

                bin_size = struct.unpack('<i', req.read(4))[0]
                if unit == self._unit and bin_size == self._resolution:
                    block_bin_count = struct.unpack('<i', req.read(4))[0]
                    block_column_count = struct.unpack('<i', req.read(4))[0]

                    n_blocks = struct.unpack('<i', req.read(4))[0]
                    for b in range(n_blocks):
                        block_number = struct.unpack('<i', req.read(4))[0]
                        file_position = struct.unpack('<q', req.read(8))[0]
                        block_size_in_bytes = struct.unpack('<i', req.read(4))[0]
                        block_map[block_number] = (file_position, block_size_in_bytes)
                else:
                    req.read(8)

                    n_blocks = struct.unpack('<i', req.read(4))[0]
                    for b in range(n_blocks):
                        req.read(16)

            if block_bin_count is None or block_column_count is None:
                raise ValueError("Matrix data for {} {} not found!".format(self._resolution, self._unit))

            region1_bins = int(region1.start / self._resolution), int(region1.end / self._resolution) + 1
            region2_bins = int(region2.start / self._resolution), int(region2.end / self._resolution) + 1

            col1, col2 = int(region1_bins[0] / block_bin_count), int(region1_bins[1] / block_bin_count)
            row1, row2 = int(region2_bins[0] / block_bin_count), int(region2_bins[1] / block_bin_count)

            blocks = set()
            for r in range(row1, row2 + 1):
                for c in range(col1, col2 + 1):
                    block_number = r * block_column_count + c
                    blocks.add(block_number)

            if region1.chromosome == region2.chromosome:
                for r in range(col1, col2 + 1):
                    for c in range(row1, row2 + 1):
                        block_number = r * block_column_count + c
                        blocks.add(block_number)

            for block_number in blocks:
                try:
                    file_position, block_size_in_bytes = block_map[block_number]

                    for x, y, weight in self._read_block(req, file_position, block_size_in_bytes):
                        if x < y:
                            if region1_bins[0] <= x < region1_bins[1] - 1 and region2_bins[0] <= y < region2_bins[1] - 1:
                                yield x + region1_chromosome_offset, y + region2_chromosome_offset, weight
                            elif region1.chromosome == region2.chromosome:
                                if region1_bins[0] <= y < region1_bins[1] - 1 and region2_bins[0] <= x < region2_bins[1] - 1:
                                    yield x + region1_chromosome_offset, y + region2_chromosome_offset, weight
                        else:
                            if region1_bins[0] <= x < region1_bins[1] - 1 and region2_bins[0] <= y < region2_bins[1] - 1:
                                yield y + region2_chromosome_offset, x + region1_chromosome_offset, weight

                except KeyError:
                    logger.debug("Could not find block {}".format(block_number))

    def _edges_subset(self, key=None, row_regions=None, col_regions=None,
                      lazy=False, *args, **kwargs):

        if row_regions[0].chromosome != row_regions[-1].chromosome:
            raise ValueError("Cannot subset rows across multiple chromosomes!")

        if col_regions[0].chromosome != col_regions[-1].chromosome:
            raise ValueError("Cannot subset columns across multiple chromosomes!")

        regions_by_ix = {}
        for region in row_regions + col_regions:
            regions_by_ix[region.ix] = region

        row_span = GenomicRegion(chromosome=row_regions[0].chromosome,
                                 start=row_regions[0].start,
                                 end=row_regions[-1].end)

        col_span = GenomicRegion(chromosome=col_regions[0].chromosome,
                                 start=col_regions[0].start,
                                 end=col_regions[-1].end)

        if not lazy:
            for x, y, weight in self._read_matrix(row_span, col_span):
                if x > y:
                    x, y = y, x
                yield Edge(source=regions_by_ix[x],
                           sink=regions_by_ix[y],
                           weight=weight)
        else:
            edge = LazyJuicerEdge(source=0, sink=0, weight=1.0, matrix=self)
            for x, y, weight in self._read_matrix(row_span, col_span):
                if x > y:
                    x, y = y, x
                edge.source, edge.sink, edge.weight = x, y, weight
                yield edge

    def _edges_iter(self, *args, **kwargs):
        chromosomes = self.chromosomes()
        for ix1 in range(len(chromosomes)):
            chromosome1 = chromosomes[ix1]
            for ix2 in range(ix1, len(chromosomes)):
                chromosome2 = chromosomes[ix2]

                key = chromosome1, chromosome2
                row_regions, col_regions = self._key_to_regions(key)

                if isinstance(row_regions, GenomicRegion):
                    row_regions = [row_regions]
                else:
                    row_regions = list(row_regions)

                if isinstance(col_regions, GenomicRegion):
                    col_regions = [col_regions]
                else:
                    col_regions = list(col_regions)

                for edge in self._edges_subset(key=key, row_regions=row_regions,
                                               col_regions=col_regions, *args, **kwargs):
                    yield edge

    @property
    def bin_size(self):
        return self._resolution

    def mappable(self, region=None):
        """
        Get the mappability vector of this matrix.
        """
        if region is None:
            if self._mappability is not None:
                return self._mappability

            mappable = []
            for chromosome, chromosome_length in self.chromosome_lengths.items():
                if chromosome.lower() == 'all':
                    continue

                norm = self.normalisation_vector(chromosome)
                for i, _ in enumerate(range(1, chromosome_length, self._resolution)):
                    if np.isnan(norm[i]):
                        mappable.append(False)
                    else:
                        mappable.append(True)

            self._mappability = mappable

            return mappable
        else:
            return np.array([r.valid for r in self.regions(region, lazy=True)])

    def bias_vector(self):
        return np.array([r.bias for r in self.regions(lazy=True)])



import numpy as np
from scipy.signal import savgol_filter
from fanc.tools.matrix import delta_window
import logging
logger = logging.getLogger(__name__)


def find_zero_crossing(x, sub_bin_precision=False):
    diff = np.diff(np.signbit(x))
    # make sure diff is not caused by nan
    diff = np.logical_and(diff, ~np.logical_or(np.isnan(x[:-1]), np.isnan(x[1:])))
    idx = np.nonzero(diff)[0]
    left, right = x[idx], x[idx + 1]
    frac = left/(left - right)
    if sub_bin_precision:
        return idx + frac
    return np.rint(idx + frac).astype(np.int_)


def getitem_interpolated(array, idx):
    if issubclass(idx.dtype.type, np.floating):
        idx_int = idx.astype(np.int_)
        frac = np.remainder(idx, 1)
        vals_left, vals_right = array[idx_int], array[idx_int + 1]
        return vals_left + frac*(vals_right - vals_left)
    return array[idx]


class BaseMaximaCaller(object):
    def __init__(self):
        self._peaks = None
        self._min_mask = None
        self._peak_mask = None
        self._scores = None

    def get_peaks(self):
        return self._peaks[self._peak_mask], self._scores[self._peak_mask]

    def get_minima(self):
        return (self._peaks[np.logical_and(self._min_mask, self._peak_mask)],
                self._scores[np.logical_and(self._min_mask, self._peak_mask)])

    def get_maxima(self):
        return (self._peaks[np.logical_and(~self._min_mask, self._peak_mask)],
                self._scores[np.logical_and(~self._min_mask, self._peak_mask)])


class MaximaCallerMatrix(BaseMaximaCaller):
    def __init__(self, x, seed_delta_ix=10, seed_delta_window=7):
        BaseMaximaCaller.__init__(self)
        self.x = x
        self.seed_delta_ix = seed_delta_ix
        self.seed_delta_window = seed_delta_window
        self._call_peaks()
        self.count = 0
        self.below = 0

    def _call_peaks(self):
        self.delta = delta_window(self.x[:, self.seed_delta_ix], self.seed_delta_window)
        self.delta_d1 = savgol_filter(self.delta, window_length=self.seed_delta_window, polyorder=2, deriv=1)
        # Figure out which delta zero crossings are minima in x
        self._peaks = np.nonzero(np.diff(np.signbit(self.delta)))[0]
        self._min_mask = self.delta_d1[self._peaks] > 0
        self._peak_mask = np.full(self._peaks.shape, True, dtype=np.bool_)

    def filter(self, min_threshold, max_threshold, start=0, end=None, lenience=.1):
        self._scores = np.ma.abs(np.ma.sum(self.x[self._peaks, start:end], axis=1))
        self.below = np.where(self._min_mask[:, np.newaxis], self.x[self._peaks] < min_threshold,
                              self.x[self._peaks] > max_threshold)
        self.count = np.sum(self.below[:, start:end], axis=1)
        pass_mask = self.count > (1 - lenience)*(end - start)
        logger.info("Discarding {}({:.1%}) of total peaks due to thresholds ({}, {})".format(np.sum(~pass_mask),
                                                                                             np.sum(~pass_mask) /
                                                                                             len(self._peaks),
                                                                                             min_threshold,
                                                                                             max_threshold))
        self._peak_mask = np.logical_and(self._peak_mask, pass_mask)


class MaximaCallerDelta(BaseMaximaCaller):
    def __init__(self, x, window_size=7, sub_bin_precision=False):
        BaseMaximaCaller.__init__(self)
        self.x = x
        self.window_size = window_size
        self.sub_bin_precision = sub_bin_precision
        self._call_peaks()

    def _call_peaks(self):
        self.delta = delta_window(self.x, self.window_size)
        self._peaks = find_zero_crossing(self.delta, sub_bin_precision=self.sub_bin_precision)
        logger.info("Found {} raw peaks".format(len(self._peaks)))
        self.delta_d1 = savgol_filter(self.delta, window_length=2*self.window_size + 1, polyorder=2, deriv=1)
        self._delta_peaks = find_zero_crossing(self.delta_d1, sub_bin_precision=self.sub_bin_precision)
        self.delta_d2 = savgol_filter(self.delta, window_length=2*self.window_size + 1, polyorder=2, deriv=2)
        # Figure out which delta zero crossings are minima in x
        self._min_mask = getitem_interpolated(self.delta_d1, self._peaks) > 0
        # Figure out which delta peaks are minima, have d2 > 0
        self._delta_min_mask = getitem_interpolated(self.delta_d2, self._delta_peaks) > 0
        # Find local extrema in delta on left and right side of x extrema
        self._left_value = np.full(self._peaks.shape, np.nan)
        self._right_value = np.full(self._peaks.shape, np.nan)
        self._right_ix = np.searchsorted(self._delta_peaks, self._peaks, side="right")
        for i, p in enumerate(self._peaks):
            try:
                self._left_value[i] = getitem_interpolated(self.delta, self._delta_peaks[self._right_ix[i] - 1])
            except IndexError:
                pass
            try:
                self._right_value[i] = getitem_interpolated(self.delta,self._delta_peaks[self._right_ix[i]])
            except IndexError:
                pass
        # Score
        self._scores = np.abs(self._left_value - self._right_value)
        self._peak_mask = np.full(self._peaks.shape, True, dtype=np.bool_)

    def filter(self, delta_score_thresh=None):
        if delta_score_thresh:
            delta_score_pass = self._scores > delta_score_thresh
            self._peak_mask = np.logical_and(self._peak_mask, delta_score_pass)
            logger.info("Discarding {}({:.1%}) of total peaks due to delta score threshold ({})".format(
                np.sum(~delta_score_pass), np.sum(~delta_score_pass)/len(self._peaks), delta_score_thresh)
            )



import argparse
import textwrap


def fancplot_parser():
    usage = '''fancplot [<fancplot global parameters>] <region> [<region> ...]
            --plot <plot type> [<plot parameters>] <plot data file(s)> [...]

            Run fancplot --plot <plot type> -h for help on a specific subplot.\n\nPlot types:\n\n'''

    command_descriptions = dict()
    for name, function in globals().items():
        if name.endswith("_parser") and name != 'fancplot_parser':
            parser = function()
            short_name = name[:-7].replace('_', '-')
            command_descriptions[short_name] = parser.description.split(".")[0]

    command_descriptions.pop('type')
    command_descriptions.pop('subplot')

    max_len = max([len(name) for name in command_descriptions.keys()]) + 4

    usage += "-- Matrix --\n"
    for name in ['triangular', 'square', 'split', 'mirror']:
        padding = ' ' * (max_len - len(name))
        usage += "{}{}{}\n".format(name, padding, command_descriptions.pop(name))

    usage += "\n-- Region --\n"
    for name in ['scores', 'line', 'bar', 'layer', 'gene']:
        padding = ' ' * (max_len - len(name))
        usage += "{}{}{}\n".format(name, padding, command_descriptions.pop(name))

    if len(command_descriptions) > 0:
        usage += "\n-- Other --\n"

    for name in command_descriptions.keys():
        padding = ' ' * (max_len - len(name))
        usage += "{}{}{}\n".format(name, padding, command_descriptions.get(name))

    parser = argparse.ArgumentParser(
        description="fancplot plotting tool for fanc",
        usage=textwrap.dedent(usage)
    )

    parser.add_argument(
        'regions',
        nargs='*',
        default=[],
        help='List of region selectors (<chr>:<start>-<end>) or '
             'files with region information (BED, GTF, ...).'
    )

    parser.add_argument(
        '-o', '--output', dest='output',
        help='Suppresses interactive plotting window and redirects plot '
             'to file. Specify path to file when plotting a single region, '
             'and path to a folder for plotting multiple regions.'
    )

    parser.add_argument(
        '-s', '--script', dest='script',
        help='Use a script file to define plot.'
    )

    parser.add_argument(
        '-p', '--plot', dest='plot',
        action='append',
        help='New plot, type will be chosen automatically by '
             'file type, unless "-t" is provided.'
    )

    parser.add_argument(
        '-n', '--name', dest='name',
        default='',
        help='Plot name to be used as prefix when plotting '
             'multiple regions. Is ignored for single region '
             'and interactive plot.'
    )

    parser.add_argument(
        '--width', dest='width',
        type=int,
        default=4,
        help='Width of the figure in inches. Default: 4'
    )

    parser.add_argument(
        '-w', '--window-size', dest='window_size',
        help='Plotting region size in base pairs. If provided, the '
             'actual size of the given region is ignored and instead '
             'a region <chromosome>:<region center - window size/2> - '
             '<region center + window size/2> will be plotted.'
    )

    parser.add_argument(
        '--invert-x', dest='invert_x',
        action='store_true',
        default=False,
        help='Invert x-axis for this plot'
    )

    parser.add_argument(
        '--tick-locations', dest='tick_locations',
        nargs='+',
        help='Manually define the locations of the tick '
             'labels on the genome axis.'
    )

    parser.add_argument(
        '-V', '--version', dest='print_version',
        action='store_true',
        default=False,
        help='Print version information'
    )
    return parser


def type_parser():
    parser = argparse.ArgumentParser(
        description="fancplot subplot identifier",
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        'type',
        help='Plot type. See fancplot -h for options.'
    )

    parser.add_argument(
        'data',
        nargs='*',
        help='Data to be plotted in subplot.'
    )

    return parser


def subplot_parser(plot_type='<plot type>'):
    parser = argparse.ArgumentParser(
        prog='fancplot <region> -p {}'.format(plot_type),
        description="fancplot subplot identifier",
    )

    parser.add_argument(
        '--aspect-ratio', dest='aspect_ratio',
        type=float,
        help='Aspect ratio of this panel. Default is determined by figure type (usually 1.0).'
    )

    parser.add_argument(
        '--title', dest='title',
        default='',
        help='Title of this plot.'
    )

    parser.add_argument(
        '--fix-chromosome', dest='fix_chromosome',
        action='store_true',
        default=False,
        help='Fix chromosome identifier for this plot '
             '(add or remove "chr" as required). Use this if'
             'there is a mismatch between the nomenclature '
             'used by different datasets in the figure, specifically '
             'if the chromosome prefix for this dataset does not match '
             'the plot region definition.'
    )

    parser.add_argument(
        '--hide-x', dest='hide_x',
        action='store_true',
        default=False,
        help='Hide x-axis for this plot'
    )

    parser.add_argument(
        '--show-minor-ticks', dest='show_minor_ticks',
        action='store_true',
        default=False,
        help='Show minor ticks on genome axis'
    )

    parser.add_argument(
        '--hide-major-ticks', dest='show_major_ticks',
        action='store_false',
        default=True,
        help='Hide major ticks on genome axis.'
    )

    parser.add_argument(
        '--show-tick-legend', dest='show_tick_legend',
        action='store_true',
        default=False,
        help='Show tick legend with distance between ticks '
             'on genome axis'
    )

    return parser


def triangular_parser():
    parser = subplot_parser('triangular')
    parser.description = 'Triangular Hi-C plot.'

    parser.add_argument(
        'hic',
        help='Hi-C object.'
    )

    parser.add_argument(
        '-vmin', '--minimum-value', dest='vmin',
        type=float,
        help='Minimum value assigned the first '
             'color in the colorbar.'
    )

    parser.add_argument(
        '-vmax', '--maximum-value', dest='vmax',
        type=float,
        help='Maximum value assigned the last color '
             'in the colorbar.'
    )

    parser.add_argument(
        '-m', '--maximum-distance', dest='max_dist',
        help='Maximum distance between two points after '
             'which triangle will be truncated.'
    )

    parser.add_argument(
        '-l', '--log', dest='log',
        action='store_true',
        default=False,
        help='Log-scale colorbar.'
    )

    parser.add_argument(
        '-r', '--range-slider', dest='adjust_range',
        action='store_true',
        default=False,
        help='Add vmax/vmin slider to plot'
    )

    parser.add_argument(
        '-u', '--uncorrected', dest='norm',
        action='store_false',
        default=True,
        help='Plot uncorrected Hi-C matrix values.'
    )

    parser.add_argument(
        '-e', '--observed-expected', dest='oe',
        action='store_true',
        default=False,
        help='Log2-O/E transform matrix values.'
    )

    parser.add_argument(
        '-c', '--colormap', dest='colormap',
        help='Matplotlib colormap'
    )

    parser.add_argument(
        '-y', '--ylabel', dest='ylabel',
        default='',
        help='Label for y axis'
    )

    parser.add_argument(
        '-d', '--default_value', dest='default_value',
        type=float,
        help='Which value to use for missing edge '
             'weights (pixels). Default: 0'
    )

    parser.add_argument(
        '-s', '--colorbar-symmetry', dest='colorbar_symmetry',
        type=float,
        default=None,
        help='Make colorbar symmetrical around this value.'
    )

    parser.add_argument(
        '--weight-field', dest='weight_field',
        help='Which value to use for plotting. '
             'Default: weight'
    )

    parser.add_argument(
        '-C', '--no-colorbar', dest='show_colorbar',
        action='store_false',
        default=True,
        help='Do not show colorbar in plot'
    )
    return parser


def square_parser():
    parser = subplot_parser('square')
    parser.description = 'Square Hi-C plot.'

    parser.add_argument(
        'hic',
        help='Hi-C object.'
    )

    parser.add_argument(
        '-vmin', '--minimum-value', dest='vmin',
        type=float,
        help='Minimum value assigned the first '
             'color in the colorbar.'
    )

    parser.add_argument(
        '-vmax', '--maximum-value', dest='vmax',
        type=float,
        help='Maximum value assigned the last '
             'color in the colorbar.'
    )

    parser.add_argument(
        '-u', '--uncorrected', dest='norm',
        action='store_false',
        default=True,
        help='Plot uncorrected Hi-C matrix values.'
    )

    parser.add_argument(
        '-e', '--observed-expected', dest='oe',
        action='store_true',
        default=False,
        help='Log2-O/E transform matrix values. '
             'Automatically sets colormap to bwr and '
             'makes colorbar symmetrical around 0.'
    )

    parser.add_argument(
        '-l', '--log', dest='log',
        action='store_true',
        default=False,
        help='Log-scale colorbar.'
    )

    parser.add_argument(
        '-r', '--range-slider', dest='adjust_range',
        action='store_true',
        default=False,
        help='Add vmax/vmin slider to plot'
    )

    parser.add_argument(
        '-f', '--flip', dest='flip',
        action='store_true',
        default=False,
        help='Flip matrix upside down'
    )

    parser.add_argument(
        '-c', '--colormap', dest='colormap',
        help='Matplotlib colormap'
    )

    parser.add_argument(
        '-s', '--colorbar-symmetry', dest='colorbar_symmetry',
        type=float,
        default=None,
        help='Make colorbar symmetrical around this value.'
    )

    parser.add_argument(
        '-C', '--no-colorbar', dest='show_colorbar',
        action='store_false',
        default=True,
        help='Do not show colorbar in plot'
    )

    parser.add_argument(
        '--weight-field', dest='weight_field',
        help='Which value to use for plotting. '
             'Default: weight'
    )
    return parser


def split_parser():
    parser = subplot_parser('split')
    parser.description = 'Matrix vs matrix plot'

    parser.add_argument(
        'hic_top',
        help='Top Hi-C object.'
    )

    parser.add_argument(
        'hic_bottom',
        help='Bottom Hi-C object.'
    )

    parser.add_argument(
        '-vmin', '--minimum-value', dest='vmin',
        type=float,
        help='Minimum value assigned the first color '
             'in the colorbar.'
    )

    parser.add_argument(
        '-vmax', '--maximum-value', dest='vmax',
        type=float,
        help='Maximum value assigned the last color '
             'in the colorbar.'
    )

    parser.add_argument(
        '-l', '--log', dest='log',
        action='store_true',
        default=False,
        help='Log-scale colorbar'
    )

    parser.add_argument(
        '-r', '--range-slider', dest='adjust_range',
        action='store_true',
        default=False,
        help='Add vmax/vmin slider to plot'
    )

    parser.add_argument(
        '-e', '--observed-expected', dest='oe',
        action='store_true',
        default=False,
        help='Log2-O/E transform matrix values. '
             'Automatically sets colormap to bwr and '
             'makes colorbar symmetrical around 0.'
    )

    parser.add_argument(
        '--colorbar-symmetry', dest='colorbar_symmetry',
        type=float,
        default=None,
        help='Make colorbar symmetrical around this value.'
    )

    parser.add_argument(
        '-c', '--colormap', dest='colormap',
        help='Matplotlib colormap'
    )

    parser.add_argument(
        '-C', '--no-colorbar', dest='show_colorbar',
        action='store_false',
        default=True,
        help='Do not show colorbar in plot'
    )

    parser.add_argument(
        '-s', '--scale-matrices', dest='scaling',
        action='store_true',
        default=False,
        help='Scale matrix values so that they sum up to the same '
             'number of contacts. Since this is potentially very '
             'time-consuming, it is disabled by default, assuming '
             'that matrices are normalised to the same total number '
             'of valid pairs'
    )
    return parser


def mirror_parser():
    parser = subplot_parser()
    parser.description = '"Mirrored" matrix comparison plot'

    parser.add_argument(
        'hic_upper',
        help='Upper Hi-C object.'
    )

    parser.add_argument(
        'hic_lower',
        help='Lower Hi-C object.'
    )

    parser.add_argument(
        '-uvmin', '--minimum-value-upper', dest='vmin_upper',
        type=float,
        help='Minimum value assigned the first color in the colorbar of upper plot.'
    )

    parser.add_argument(
        '-uvmax', '--maximum-value-upper', dest='vmax_upper',
        type=float,
        help='Maximum value assigned the last color in the colorbar of upper plot.'
    )

    parser.add_argument(
        '-lvmin', '--minimum-value-lower', dest='vmin_lower',
        type=float,
        help='Minimum value assigned the first color in '
             'the colorbar of lower plot.'
    )

    parser.add_argument(
        '-lvmax', '--maximum-value-lower', dest='vmax_lower',
        type=float,
        help='Maximum value assigned the last color in '
             'the colorbar of lower plot.'
    )

    parser.add_argument(
        '-d', '--maximum-distance', dest='max_dist',
        help='Maximum distance between two points after '
             'which triangle will be truncated.'
    )

    parser.add_argument(
        '-ul', '--log-upper', dest='log_upper',
        action='store_true',
        default=False,
        help='Log-scale colorbar for upper plot'
    )

    parser.add_argument(
        '-ll', '--log-lower', dest='log_lower',
        action='store_true',
        default=False,
        help='Log-scale colorbar for lower plot'
    )

    parser.add_argument(
        '-uc', '--colormap-upper', dest='colormap_upper',
        help='Matplotlib colormap for upper plot'
    )

    parser.add_argument(
        '-lc', '--colormap-lower', dest='colormap_lower',
        help='Matplotlib colormap for lower plot'
    )

    parser.add_argument(
        '-ue', '--observed-expected-upper', dest='oe_upper',
        action='store_true',
        default=False,
        help='Log2-O/E transform matrix values of upper matrix. '
             'Automatically sets colormap to bwr and '
             'makes colorbar symmetrical around 0.'
    )

    parser.add_argument(
        '-le', '--observed-expected-lower', dest='oe_lower',
        action='store_true',
        default=False,
        help='Log2-O/E transform matrix values of lower matrix. '
             'Automatically sets colormap to bwr and '
             'makes colorbar symmetrical around 0.'
    )

    parser.add_argument(
        '--colorbar-symmetry-upper', dest='colorbar_symmetry_upper',
        type=float,
        default=None,
        help='Make upper colorbar symmetrical around this value.'
    )

    parser.add_argument(
        '--colorbar-symmetry-lower', dest='colorbar_symmetry_lower',
        type=float,
        default=None,
        help='Make lower colorbar symmetrical around this value.'
    )

    parser.add_argument(
        '-C', '--no-colorbars', dest='show_colorbar',
        action='store_false',
        default=True,
        help='Do not show colorbars in plot'
    )
    return parser


def scores_parser():
    parser = subplot_parser()
    parser.description = 'Region scores plot with parameter dependency.'

    parser.add_argument(
        'scores',
        help='Array object, e.g. InsulationScores, DirectionalityIndexes, ... .'
    )

    parser.add_argument(
        '-vmin', '--minimum-value', dest='vmin',
        type=float,
        help='Minimum value assigned the first color in the colorbar.'
    )

    parser.add_argument(
        '-vmax', '--maximum-value', dest='vmax',
        type=float,
        help='Maximum value assigned the last color in the colorbar.'
    )

    parser.add_argument(
        '-l', '--log', dest='log',
        action='store_true',
        default=False,
        help='Log-transform heatmap values'
    )

    parser.add_argument(
        '-r', '--range', dest='range',
        nargs=2,
        type=int,
        help='Range of y-values to plot (<min> <max> inclusive)'
    )

    parser.add_argument(
        '-p', '--parameters', dest='parameters',
        type=str,
        nargs='+',
        help='List of specific window sizes / parameters to plot.'
    )

    parser.add_argument(
        '-c', '--colormap', dest='colormap',
        default='RdBu_r',
        help='Matplotlib colormap (default: RdBu_r)'
    )

    parser.add_argument(
        '-C', '--no-colorbar', dest='show_colorbar',
        action='store_false',
        default=True,
        help='Do not show colorbar in plot'
    )

    parser.add_argument(
        '-S', '--no-symmetry', dest='symmetry',
        action='store_false',
        default=True,
        help='Do not plot colormap symmetrical around 0.'
    )

    parser.add_argument(
        '-g', '--genomic-format', dest='genomic_format',
        action='store_true',
        default=False,
        help='Use abbreviated genomic formatting for y axis '
             'labels (e.g. 1kb instead of 1000).'
    )

    parser.add_argument(
        '-rc', '--replacement-color', dest='replacement_color',
        default='grey',
        help='Color to replace missing values. Default: grey'
    )
    return parser


def line_parser():
    parser = subplot_parser()
    parser.description = '''Line plot.'''

    parser.add_argument(
        'regions',
        nargs='+',
        help='Region-based file, e.g. BED, GFF, BigWig, ...'
    )

    parser.add_argument(
        '-a', '--attribute', dest='attribute',
        default='score',
        help='Attribute to plot. Default: score'
    )

    parser.add_argument(
        '-b', '--bin-size', dest='bin_size',
        type=int,
        help='Bin size. Specify if you want to bin scores into '
             'larger regions (using weighted mean).'
    )

    parser.add_argument(
        '-l', '--labels', dest='labels',
        nargs='+',
        help='Labels for region datasets'
    )

    parser.add_argument(
        '-c', '--colors', dest='colors',
        nargs='+',
        help='Colors for region datasets'
    )

    parser.add_argument(
        '-f', '--fill', dest='fill',
        action='store_true',
        default=False,
        help='Fill region between the line and the x axis.'
    )

    parser.add_argument(
        '-s', '--line-style', dest='line_style',
        default='mid',
        help='Style of line. Default is "mid": Connect regions only at their'
             'midpoint. Alternatively use "step", where the whole region '
             'is assigned a value, leading to rectangular appearance. '
    )

    parser.add_argument(
        '-y', '--ylim', dest='ylim',
        nargs=2,
        type=float,
        help='''Y-axis limits.'''
    )

    parser.add_argument(
        '--alpha', dest='alpha',
        default=0.5,
        type=float,
        help='Transparency of bars. Value between 0 and 1, default: 0.5'
    )

    parser.add_argument(
        '--legend-location', dest='legend_location',
        default='best',
        help='Location of the legend when providing data labels.'
    )
    return parser


def bar_parser():
    parser = subplot_parser()
    parser.description = '''Bar plot for region scores.'''

    parser.add_argument(
        'regions',
        nargs='+',
        help='Region-based file, e.g. BED, GFF, BigWig, ...'
    )

    parser.add_argument(
        '-a', '--attribute', dest='attribute',
        default='score',
        help='Attribute to plot. Default: score'
    )

    parser.add_argument(
        '--alpha', dest='alpha',
        default=0.5,
        type=float,
        help='Transparency of bars. Value between 0 and 1, default: 0.5'
    )

    parser.add_argument(
        '-l', '--labels', dest='labels',
        nargs='+',
        help='Labels for region datasets'
    )

    parser.add_argument(
        '-y', '--ylim', dest='ylim',
        nargs=2,
        type=float,
        help='''Y-axis limits.'''
    )

    parser.add_argument(
        '-c', '--colors', dest='colors',
        nargs='+',
        help='Colors for region datasets'
    )

    parser.add_argument(
        '--legend-location', dest='legend_location',
        default='best',
        help='Location of the legend when providing data labels.'
    )

    return parser


def gene_parser():
    parser = subplot_parser()
    parser.description = '''Gene plot.'''

    parser.add_argument(
        'genes',
        help='''Genes file (GTF, BED).'''
    )

    parser.add_argument(
        '-f', '--feature-types', dest='feature_types',
        nargs='+',
        default=('exon',),
        help='''Feature types to be plotted. By default only exons are shown.'''
    )

    parser.add_argument(
        '-cn', '--color-neutral', dest='color_neutral',
        default='gray',
        help='''Color for genes without strand information.'''
    )

    parser.add_argument(
        '-cf', '--color-forward', dest='color_forward',
        default='orangered',
        help='''Color for genes on the '+' strand.'''
    )

    parser.add_argument(
        '-cr', '--color-reverse', dest='color_reverse',
        default='darkturquoise',
        help='''Color for genes on the '-' strand.'''
    )

    parser.add_argument(
        '-cs', '--color-score', dest='color_score',
        action='store_true',
        default=False,
        help='''Assign color based on score value.'''
    )

    parser.add_argument(
        '-b', '--box-height', dest='box_height',
        type=float,
        default=0.1,
        help='''Height of exon boxes. Default: 0.1'''
    )

    parser.add_argument(
        '-s', '--font-size', dest='font_size',
        type=float,
        default=9.,
        help='''Font size for gene labels. Default: 9'''
    )

    parser.add_argument(
        '-a', '--arrow-size', dest='arrow_size',
        type=float,
        default=8.,
        help='''Size of directionality arrows. Default: 8'''
    )

    parser.add_argument(
        '-l', '--line-width', dest='line_width',
        type=float,
        default=1.,
        help='''Width of the line along the length of the gene. Default: 2.'''
    )

    parser.add_argument(
        '-g', '--group-by', dest='group_by',
        default='transcript_id',
        help='''Group exons according to this attribute (e.g. for
                            plotting multiple transcripts of the same gene.
                            Default: transcript_id'''
    )

    parser.add_argument(
        '--label-field', dest='label_field',
        default='name',
        help='''Which field to use as label.'''
    )

    parser.add_argument(
        '-m', '--min-gene-size', dest='min_gene_size',
        type=int,
        help='''If set, forces the plotted gene to be at least this wide.'''
    )

    parser.add_argument(
        '-L', '--no-labels', dest='show_labels',
        action='store_false',
        help='''Do not plot labels on genes.'''
    )
    parser.set_defaults(show_labels=True)

    parser.add_argument(
        '-C', '--collapse', dest='collapse',
        action='store_true',
        help='''Collapse all genes onto a single row.'''
    )
    parser.set_defaults(collapse=False)

    parser.add_argument(
        '-sq', '--squash', dest='squash',
        action='store_true',
        help='''Merge all exons of each grouping unit together.
                Useful especially when setting --group-by to "gene_id" or "gene_symbol".
                Overlapping genes will still draw on separate rows in contrast to --collapse'''
    )
    parser.set_defaults(squash=False)
    return parser


def layer_parser():
    parser = subplot_parser()
    parser.description = 'Layered feature plot.'

    parser.add_argument(
        'features',
        help='Features file (GFF, BED, Tabix).'
    )

    parser.add_argument(
        '-g', '--grouping', dest='grouping_attribute',
        help='GFF attribute to use for grouping into layers. For BED this is always the name column.'
    )

    parser.add_argument(
        '-i', '--include', dest='include',
        nargs='+',
        help='Include only these groups.'
    )

    parser.add_argument(
        '-e', '--exclude', dest='exclude',
        nargs='+',
        help='Exclude these groups.'
    )

    parser.add_argument(
        '-S', '--no-shadow', dest='shadow',
        action='store_false',
        default=True,
        help='Plot element size "as is" without surrounding box to make it more visible.'
    )

    parser.add_argument(
        '-C', '--collapse', dest='collapse',
        action='store_true',
        default=False,
        help='Plot all elements in one row.'
    )

    parser.add_argument(
        '-w', '--shadow-width', dest='shadow_width',
        type=float,
        default=0.005,
        help='Vertical distance between rows of genes in the plot.'
    )

    parser.add_argument(
        '--no-colors', dest='colors',
        action='store_false',
        default=True,
        help='Do not add color to plots.'
    )

    return parser


cdef extern from  "natural_cmp.c":
    int strnum_cmp(const char *_a, const char *_b)

slash_int = b'/'[0]

def natural_cmp(a, b, ignore_pe_suffix=True):
    assert (isinstance(a, bytes))
    assert (isinstance(b, bytes))

    if ignore_pe_suffix and len(a) > 1 and len(b) > 1 and a[-2] == b[-2] == slash_int:
        return strnum_cmp(a[:-2], b[:-2])
    else:
        return strnum_cmp(a, b)



import genomic_regions as gr
import numpy as np
import logging

logger = logging.getLogger(__name__)


def load_contacts(contacts_file_name, sep=None, ix_converter=None):
    m = None
    try:  # numpy binary format
        m = np.load(contacts_file_name)
    except (IOError, ValueError):  # not an .npy file

        # check number of fields in file
        with open(contacts_file_name, 'r') as f:
            line = f.readline()
            while line.startswith('#'):
                line = f.readline()
            line = line.rstrip()
            n_fields = len(line.split(sep))

        if n_fields > 3:  # square matrix format
            m = np.loadtxt(contacts_file_name)

        if m is not None:
            for i, row in enumerate(m):
                for j, val in enumerate(row):
                    yield i, j, val
        else:
            with open(contacts_file_name, 'r') as f:
                for line in f:
                    if line.startswith("#"):
                        continue
                    line = line.rstrip()
                    fields = line.split(sep)
                    if ix_converter is None:
                        source, sink, weight = int(fields[0]), int(fields[1]), float(fields[2])
                    else:
                        source = ix_converter[fields[0]]
                        sink = ix_converter[fields[1]]
                        weight = float(fields[2])

                    yield source, sink, weight


def load_regions(file_name, sep=None):
    regions = []
    ix_converter = None
    with open(file_name, 'r') as f:
        for i, line in enumerate(f):
            line = line.rstrip()
            fields = line.split(sep)
            if len(fields) > 2:
                chromosome = fields[0]
                start = int(fields[1])
                end = int(fields[2])
                ix = i
                if len(fields) > 3 and fields[3] != '.':  # HicPro
                    if ix_converter is None:
                        ix_converter = dict()
                    if fields[3] in ix_converter:
                        raise ValueError("Name column in region BED must "
                                         "only contain unique values! ({})".format(fields[3]))
                    ix_converter[fields[3]] = ix
                regions.append(gr.GenomicRegion(chromosome=chromosome, start=start, end=end, ix=ix))
    return regions, ix_converter



from __future__ import division

import tables
import logging
import operator
from collections import defaultdict
from genomic_regions import GenomicRegion, as_region
from future.utils import string_types

from ..tools.general import RareUpdateProgressBar
from ..regions import RegionsTable
from ..matrix import RegionMatrixTable, RegionMatrix
from .domains import RegionScoreParameterTable

import numpy as np
from sklearn.decomposition import PCA

logger = logging.getLogger(__name__)


def _edge_collection(*hics, region=None, scale=True,
                     filters=None, **kwargs):
    """
    Get all weights from the same edge across different Hi-C matrices.

    :param hics: Hic/matrix objects or list of Hic/Matrix objects
    :param region: Optional region subset (e.g. a single chromosome)
    :param scale: True if Hic matrices should be scaled to the
                  number of valid pairs
    :return: dict of lists
    """
    inter_chromosomal = kwargs.pop("inter_chromosomal", False)

    if len(hics) == 1:
        hics = hics[0]

    if filters is None:
        filters = []

    scaling_factors = [1.0] * len(hics)
    if scale:
        reference_hic = None
        for i, hic in enumerate(hics):
            if i == 0:
                reference_hic = hic
            else:
                s = hic.scaling_factor(reference_hic)
                scaling_factors[i] = s

    total_edges = sum(len(hic.edges) for hic in hics)

    if isinstance(region, GenomicRegion) or isinstance(region, string_types) or region is None:
        regions = [region]
    else:
        regions = region

    region_pairs = []
    for i in range(len(regions)):
        region1 = as_region(regions[i]) if regions[i] is not None else None
        if region1 is None:
            region_pairs.append(None)
            continue

        for j in range(i, len(regions)):
            region2 = as_region(regions[j]) if regions[j] is not None else None
            if not inter_chromosomal and region1.chromosome != region2.chromosome:
                continue
            region_pairs.append((region1, region2))

    edges = defaultdict(list)
    j = 0
    with RareUpdateProgressBar(max_value=total_edges, prefix='Edge collection') as pb:
        for i, hic in enumerate(hics):
            logger.debug("Adding Hic {} ({}) to edge collection".format(i, region))

            for region_pair in region_pairs:
                for edge in hic.edges(region_pair, lazy=True,
                                      inter_chromosomal=inter_chromosomal, **kwargs):
                    weight = edge.weight * scaling_factors[i]
                    source, sink = edge.source, edge.sink

                    weight_list = edges[(source, sink)]
                    while len(weight_list) < i:
                        weight_list.append(0)

                    if len(weight_list) <= i:
                        weight_list.append(weight)

                    pb.update(j)
                    j += 1

    total = len(hics)
    for _, weights in edges.items():
        while len(weights) < total:
            weights.append(0)

    before_filtering = len(edges)
    for key in list(edges.keys()):
        for f in filters:
            if not f.valid(key[0], key[1], edges[key]):
                del edges[key]
    after_filtering = len(edges)

    logger.info("Valid edges: {}/{}".format(after_filtering, before_filtering))

    return edges


class EdgeCollectionFilter(object):
    def __init__(self):
        pass

    def valid(self, source, sink, weights):
        raise NotImplementedError("Subclasses of EdgeCollectionFilter must implement "
                                  "'valid'!")


class ObservedExpectedFilter(EdgeCollectionFilter):
    def __init__(self, hics, fold_change_cutoff=1.0, min_valid_edges=1,
                 oe_per_chromosome=True):
        EdgeCollectionFilter.__init__(self)

        self._cutoff = fold_change_cutoff
        self._min_valid = min_valid_edges
        self._oe_per_chromosome = oe_per_chromosome

        self._intra_expected = {}
        self._inter_expected = {}
        self._intra_expected_chromosome = {}
        for i, hic in enumerate(hics):
            intra_expected, intra_expected_chromosome, inter_expected = hic.expected_values()
            self._intra_expected[i] = intra_expected
            self._inter_expected[i] = inter_expected
            self._intra_expected_chromosome[i] = intra_expected_chromosome

        self._chromosome_dict = {}
        for region in hics[0].regions:
            self._chromosome_dict[region.ix] = region.chromosome

    def valid(self, source, sink, weights):
        source_chromosome = self._chromosome_dict[source]
        sink_chromosome = self._chromosome_dict[sink]

        count_valid = 0
        for i, weight in enumerate(weights):
            if source_chromosome == sink_chromosome:
                d = abs(sink - source)
                if self._oe_per_chromosome:
                    e = self._intra_expected_chromosome[i][source_chromosome][d]
                else:
                    e = self._intra_expected[i][d]
            else:
                e = self._inter_expected[i]

            fc = weight / e
            if np.isfinite(fc) and fc > self._cutoff:
                count_valid += 1

        if count_valid > self._min_valid:
            return True
        return False


class NonzeroFilter(EdgeCollectionFilter):
    def __init__(self):
        EdgeCollectionFilter.__init__(self)

    def valid(self, source, sink, weights):
        if 0 in weights:
            return False
        return True


class AbsoluteWeightFilter(EdgeCollectionFilter):
    def __init__(self, cutoff, min_above_cutoff=1):
        EdgeCollectionFilter.__init__(self)

        self.cutoff = cutoff
        self._min_valid = min_above_cutoff

    def valid(self, source, sink, weights):
        counter = 0
        for weight in weights:
            if weight >= self.cutoff:
                counter += 1

        if counter < self._min_valid:
            return False
        return True


class MinMaxDistanceFilter(EdgeCollectionFilter):
    def __init__(self, hics, min_distance=None, max_distance=None):
        self.bin_size = None
        for hic in hics:
            if self.bin_size is None:
                self.bin_size = hic.bin_size
            else:
                if self.bin_size != hic.bin_size:
                    raise ValueError("Hic objects must have same bin size! "
                                     "({}/{})".format(self.bin_size, hic.bin_size))
        EdgeCollectionFilter.__init__(self)
        self.min_distance = np.round(min_distance/self.bin_size) if min_distance is not None else None
        self.max_distance = np.round(max_distance/self.bin_size) if max_distance is not None else None

    def valid(self, source, sink, weights):
        d = abs(sink - source)
        if self.min_distance is not None and d < self.min_distance:
            return False
        if self.max_distance is not None and d > self.max_distance:
            return False
        return True


class ComparisonMatrix(RegionMatrixTable):

    _classid = 'COMPARISONMATRIX'

    def __init__(self, *args, **kwargs):
        RegionMatrixTable.__init__(self, *args, **kwargs)

    def compare(self, weight1, weight2):
        """
        Compare two edge weights.

        :param weight1: float
        :param weight2: float
        :return: float
        """
        raise NotImplementedError("Subclasses of ComparisonMatrix must implement "
                                  "'compare'")

    @classmethod
    def from_matrices(cls, matrix1, matrix2, file_name=None, tmpdir=None,
                      log_cmp=False, ignore_infinite=True, ignore_zeros=False,
                      scale=True, **kwargs):
        kwargs['lazy'] = True

        comparison_matrix = cls(file_name=file_name, mode='w', tmpdir=tmpdir)
        comparison_matrix.add_regions(matrix1.regions, preserve_attributes=False)

        sf = 1.0
        if scale:
            if kwargs.get('oe', False):
                logger.warning('Not computing scaling factor due to O/E conversion.')
            else:
                sf = matrix2.scaling_factor(matrix1)

        compare = comparison_matrix.compare
        chromosomes = matrix1.chromosomes()
        n_chromosome_pairs = int(np.round(len(chromosomes)**2/2 + len(chromosomes)/2))
        current_chromosome_pair = 0
        with RareUpdateProgressBar(max_value=n_chromosome_pairs, prefix='Compare') as pb:
            for chr_i in range(len(chromosomes)):
                chromosome1 = chromosomes[chr_i]
                for chr_j in range(chr_i, len(chromosomes)):
                    chromosome2 = chromosomes[chr_j]

                    edges1 = dict()
                    for edge in matrix1.edges((chromosome1, chromosome2), **kwargs):
                        edges1[(edge.source, edge.sink)] = edge.weight

                    edges2 = dict()
                    for edge in matrix2.edges((chromosome1, chromosome2), **kwargs):
                        edges2[(edge.source, edge.sink)] = edge.weight * sf

                    for key, w1 in edges1.items():
                        try:
                            w2 = edges2[key]
                        except KeyError:
                            w2 = matrix2._default_value

                        if ignore_zeros and (w1 == 0 or w2 == 0):
                            continue
                        weight = compare(w1, w2)
                        if log_cmp:
                            weight = np.log2(weight)
                        if ignore_infinite and not np.isfinite(weight):
                            continue

                        comparison_matrix.add_edge_simple(key[0], key[1], weight=weight)

                    for key, w2 in edges2.items():
                        if key not in edges1:
                            w1 = matrix1._default_value

                            if ignore_zeros and (w1 == 0 or w2 == 0):
                                continue
                            weight = compare(w1, w2)
                            if log_cmp:
                                weight = np.log2(weight)
                            if ignore_infinite and not np.isfinite(weight):
                                continue

                            comparison_matrix.add_edge_simple(key[0], key[1], weight=weight)

                    current_chromosome_pair += 1
                    pb.update(current_chromosome_pair)

        comparison_matrix.flush()
        return comparison_matrix


class FoldChangeMatrix(ComparisonMatrix):

    _classid = 'FOLDCHANGEMATRIX'

    def __init__(self, *args, **kwargs):
        ComparisonMatrix.__init__(self, *args, **kwargs)

    def compare(self, weight1, weight2):
        try:
            return weight1 / weight2
        except ZeroDivisionError:
            return np.nan


class DifferenceMatrix(ComparisonMatrix):

    _classid = 'DIFFERENCEMATRIX'

    def __init__(self, *args, **kwargs):
        ComparisonMatrix.__init__(self, *args, **kwargs)

    def compare(self, weight1, weight2):
        return weight1 - weight2


class ComparisonScores(RegionScoreParameterTable):

    _classid = 'COMPARISONSCORES'

    def __init__(self, *args, **kwargs):
        RegionScoreParameterTable.__init__(self, *args, **kwargs)

    def compare(self, score1, score2):
        """
        Compare two edge weights.

        :param score1: float
        :param score2: float
        :return: float
        """
        raise NotImplementedError("Subclasses of ComparisonScores must implement "
                                  "'compare'")

    @classmethod
    def from_scores(cls, scores1, scores2, attributes=None, file_name=None, tmpdir=None,
                    log=False, field_prefix='cmp_', *args, **kwargs):
        # all matching parameters
        if attributes is None:
            attributes = []
            attributes1 = scores1._score_fields
            for a, p in zip(scores2._score_fields, scores2._parameters):
                if a in attributes1:
                    attributes.append(p)

        comparison_scores = cls(parameter_values=attributes, parameter_prefix=field_prefix,
                                file_name=file_name, mode='w', tmpdir=tmpdir)
        comparison_scores.add_regions(scores1.regions, preserve_attributes=False)

        region_pairs = list()
        region_ixs = dict()
        for region in scores1.regions:
            region_ixs[(region.chromosome, region.start, region.end)] = len(region_pairs)
            region_pairs.append([region, None])

        for region in scores2.regions:
            ix = region_ixs[(region.chromosome, region.start, region.end)]
            region_pairs[ix][1] = region

        for attribute in attributes:
            field = scores1._score_field_converter([attribute])[0]
            cmp_scores = []
            for r1, r2 in region_pairs:
                v1 = getattr(r1, field)
                v2 = getattr(r2, field)
                v_cmp = comparison_scores.compare(v1, v2)
                if log:
                    v_cmp = np.log2(v_cmp)
                cmp_scores.append(v_cmp)
            comparison_scores.scores(attribute, cmp_scores)

        return comparison_scores


class FoldChangeScores(ComparisonScores):

    _classid = 'FOLDCHANGESCORES'

    def __init__(self, *args, **kwargs):
        ComparisonScores.__init__(self, *args, **kwargs)

    def compare(self, score1, score2):
        return score1 / score2


class DifferenceScores(ComparisonScores):

    _classid = 'DIFFERENCESCORES'

    def __init__(self, *args, **kwargs):
        ComparisonScores.__init__(self, *args, **kwargs)

    def compare(self, score1, score2):
        return score1 - score2


class ComparisonRegions(RegionsTable):

    _classid = 'COMPARISONREGIONS'

    def __init__(self, *args, **kwargs):
        RegionsTable.__init__(self, *args, **kwargs)

    def compare(self, score1, score2):
        """
        Compare two edge weights.

        :param score1: float
        :param score2: float
        :return: float
        """
        raise NotImplementedError("Subclasses of ComparisonRegions must implement "
                                  "'compare'")

    @classmethod
    def from_regions(cls, region_based1, region_based2, attribute='score',
                     file_name=None, tmpdir=None, log=False, score_field='score',
                     *args, **kwargs):
        comparison_regions = cls(file_name=file_name, mode='w', tmpdir=tmpdir,
                                 additional_fields={attribute: tables.Float32Col()})
        comparison_regions.add_regions(region_based1.regions, preserve_attributes=False)

        regions = dict()
        for region in region_based1.regions:
            regions[(region.chromosome, region.start, region.end)] = region

        scores = []
        for region2 in region_based2.regions:
            region1 = regions[(region2.chromosome, region2.start, region2.end)]
            v1 = getattr(region1, attribute)
            v2 = getattr(region2, attribute)
            v = comparison_regions.compare(v1, v2)
            if log:
                v = np.log2(v)
            scores.append(v)
        comparison_regions.region_data(score_field, scores)
        return comparison_regions


class FoldChangeRegions(ComparisonRegions):

    _classid = 'FOLDCHANGEREGIONS'

    def __init__(self, *args, **kwargs):
        ComparisonRegions.__init__(self, *args, **kwargs)

    def compare(self, score1, score2):
        return score1 / score2


class DifferenceRegions(ComparisonRegions):

    _classid = 'DIFFERENCEREGIONS'

    def __init__(self, *args, **kwargs):
        ComparisonRegions.__init__(self, *args, **kwargs)

    def compare(self, score1, score2):
        return score1 - score2


class SplitMatrix(object):
    def __init__(self, matrix_top, matrix_bottom, scaling_factor=1.):
        self.matrix_top = matrix_top
        self.matrix_bottom = matrix_bottom
        self.scaling_factor = scaling_factor

    def matrix(self, *args, **kwargs):
        """
        Return a :class:`~HicMatrix` where values above the diagonal
        are from this object and values below the diagonal are from
        another :class:`~Hic` object.

        "Above the diagonal" refers to the diagonal of the complete
        Hic object, not the diagonal of the returned matrix.

        :param key: A matrix selector. Use tuple to select row and
                    columns
        :param scaling_factor: Factor to scale the hic values. If None,
                               will be computed using
                               :func:`~Hic.scaling_factor`.
        :param kwargs: Keyword arguments passed to both matrices
                       :func:`~fanc.matrix.RegionMatrixContainer.matrix`
                       functions.
        :return: :class:`~RegionMatrix`
        """
        m_top = self.matrix_top.matrix(*args, **kwargs)

        # find diagonal
        row_region = m_top.row_regions[0]
        matching_index = None
        for i, col_region in enumerate(m_top.col_regions):
            if col_region == row_region:
                matching_index = i

        if matching_index is None:
            col_region = m_top.col_regions[0]
            for i, row_region in enumerate(m_top.row_regions):
                if col_region == row_region:
                    matching_index = -1 * i

        if matching_index is None:
            return m_top

        # replace diagonal
        m_bottom = self.matrix_bottom.matrix(*args, **kwargs) * self.scaling_factor
        top_indices = np.triu_indices(m_top.shape[0], matching_index, m_top.shape[1])
        m_bottom[top_indices] = m_top[top_indices]

        return RegionMatrix(m_bottom, row_regions=m_top.row_regions, col_regions=m_top.col_regions)


class EdgeCollectionSelector(object):
    def __init__(self):
        pass

    def filter_edge_collection(self, edge_collection, sample_size, *args, **kwargs):
        raise NotImplementedError("Subclasses of EdgeCollectionSelector must implement "
                                  "filter_edge_collection!")


class LargestVarianceSelector(EdgeCollectionSelector):
    def __init__(self):
        EdgeCollectionSelector.__init__(self)

    def filter_edge_collection(self, edge_collection, sample_size, *args, **kwargs):
        if sample_size is None:
            sample_size = len(edge_collection)

        variances = dict()
        for key, weights in edge_collection.items():
            v = np.nanvar(weights)
            if np.isfinite(v):
                variances[key] = v

        variances_sorted = sorted(variances.items(), key=operator.itemgetter(1), reverse=True)

        for i in range(min(len(variances_sorted), sample_size)):
            source, sink = variances_sorted[i][0]
            weights = edge_collection[(source, sink)]
            yield source, sink, weights


class PassthroughSelector(EdgeCollectionSelector):
    def __init__(self):
        EdgeCollectionSelector.__init__(self)

    def filter_edge_collection(self, edge_collection, sample_size, *args, **kwargs):
        if sample_size is None:
            sample_size = len(edge_collection)

        for i, (key, weights) in enumerate(edge_collection.items()):
            if i > sample_size:
                break
            yield key[0], key[1], weights


class LargestFoldChangeSelector(EdgeCollectionSelector):
    def __init__(self):
        EdgeCollectionSelector.__init__(self)

    def filter_edge_collection(self, edge_collection, sample_size=None, *args, **kwargs):
        if sample_size is None:
            sample_size = len(edge_collection)

        max_fold_changes = dict()
        for key, weights in edge_collection.items():
            min_weight = np.nanmin(weights)
            max_weight = np.nanmax(weights)

            fc = max_weight / min_weight
            if np.isfinite(fc):
                max_fold_changes[key] = fc

        fc_sorted = sorted(max_fold_changes.items(), key=operator.itemgetter(1), reverse=True)

        for i in range(min(len(fc_sorted), sample_size)):
            source, sink = fc_sorted[i][0]
            weights = edge_collection[(source, sink)]
            yield source, sink, weights


def hic_pca(*hics, sample_size=None, region=None, strategy='variance', scale=True, log=False,
            ignore_zeros=False, oe_enrichment=None, min_distance=None, max_distance=None,
            background_ligation=False, min_libraries_above_background=1,
            **kwargs):

    strategies = {
        'variance': LargestVarianceSelector(),
        'fold-change': LargestFoldChangeSelector(),
        'passthrough': PassthroughSelector(),
        None: PassthroughSelector(),
    }

    selector = strategies[strategy]

    filters = []
    if ignore_zeros:
        filters.append(NonzeroFilter())
    if oe_enrichment is not None:
        filters.append(ObservedExpectedFilter(hics,
                                              fold_change_cutoff=oe_enrichment,
                                              min_valid_edges=min_libraries_above_background))
    if min_distance is not None or max_distance is not None:
        filters.append(MinMaxDistanceFilter(hics, min_distance=min_distance, max_distance=max_distance))
    if background_ligation:
        hic = hics[0]
        _, _, inter_expected = hic.expected_values()
        filters.append(AbsoluteWeightFilter(cutoff=inter_expected,
                                            min_above_cutoff=min_libraries_above_background))

    edge_collection = _edge_collection(*hics, region=region, scale=scale,
                                       filters=filters, **kwargs)

    pca_edges = selector.filter_edge_collection(edge_collection, sample_size)

    values = []
    for source, sink, weights in pca_edges:
        values.append(weights)

    if not log:
        y = np.array(values)
    else:
        y = np.log(np.array(values))

    pca = PCA()
    pca_res = pca.fit_transform(y.T)
    logger.info("Variance explained: %s" % str(pca.explained_variance_ratio_))

    return pca, pca_res



"""
Provide basic convenience data types.

Most classes are based on pytables and hdf5 dictionaries,
allowing on-disk storage and therefore processing of large
files. Other features include indexing and querying.
"""

from __future__ import division
from fanc.config import config
import tables as t
from tables.nodes import filenode
from fanc.tools.files import create_or_open_pytables_file, tmp_file_name
from fanc.tools.general import RareUpdateProgressBar, create_col_index
import os
from tables.exceptions import NoSuchNodeError
from abc import ABCMeta, abstractmethod
import shutil
import warnings
from collections import defaultdict
from .registry import class_id_dict, class_name_dict
import tempfile
from future.utils import with_metaclass, string_types
from builtins import object
import logging
logger = logging.getLogger(__name__)

_filter = t.Filters(complib="blosc", complevel=2, shuffle=True)


class MetaFileBased(type):
    """
    Metaclass that exists to register classes by name in the module.
    Taken pretty much verbatim from the same system in PyTables.
    """
    def __init__(cls, name, bases, dict_):
        super(MetaFileBased, cls).__init__(name, bases, dict_)

        # Always register into class name dictionary.
        class_name_dict[cls.__name__] = cls

        # Register into class identifier dictionary only if the class
        # has an identifier and it is different from its parents'.
        cid = getattr(cls, '_classid', None)
        if cid is not None:
            for base in bases:
                pcid = getattr(base, '_classid', None)
                if pcid == cid:
                    break
            else:
                class_id_dict[cid] = cls


class FileBased(with_metaclass(MetaFileBased, object)):
    _classid = 'FILEBASED'

    def __init__(self, file_name=None, mode='a', tmpdir=None,
                 _meta_group='meta_information'):

        if hasattr(self, 'tmp_file_name'):
            self.tmp_file_name = getattr(self, 'tmp_file_name')
        else:
            self.tmp_file_name = None

        # open file or keep in memory
        if hasattr(self, 'file'):
            if not isinstance(self.file, t.file.File):
                raise ValueError("'file' attribute already exists, but is no pytables File")
        else:
            self.file = None
            self.file_name = file_name

            if tmpdir is None or (isinstance(tmpdir, bool) and not tmpdir):
                self.tmp_file_name = None
                self._init_file(file_name, mode)
            else:
                logger.info("Working in temporary directory...")
                if isinstance(tmpdir, bool):
                    tmpdir = tempfile.gettempdir()
                else:
                    tmpdir = os.path.expanduser(tmpdir)
                self.tmp_file_name = tmp_file_name(tmpdir, prefix='tmp_fanc', extension='h5')
                logger.info("Temporary output file: {}".format(self.tmp_file_name))
                if mode in ('r+', 'r'):
                    shutil.copyfile(file_name, self.tmp_file_name)
                elif mode == 'a' and file_name is not None and os.path.isfile(file_name):
                    shutil.copyfile(file_name, self.tmp_file_name)
                self._init_file(self.tmp_file_name, mode)

        if not hasattr(self, 'meta'):
            self._meta_group_name = _meta_group
            self.meta = None
            self._init_meta()
            self._update_classid()

    def _init_meta(self):
        class MetaAccess(object):
            def __init__(self, meta_attributes=None):
                self._meta_attributes = meta_attributes

            def __getitem__(self, item):
                if self._meta_attributes is None:
                    raise KeyError("No such key: {}", item)
                return self._meta_attributes[item]

            def __getattr__(self, item):
                if item == '_meta_attributes':
                    return object.__getattribute__(self, item)
                if self._meta_attributes is None:
                    raise AttributeError("No such attribute: {}", item)
                return getattr(self._meta_attributes, item)

            def __setattr__(self, key, value):
                if key == '_meta_attributes':
                    object.__setattr__(self, key, value)
                    return

                if self._meta_attributes is None:
                    raise t.FileModeError("File not writable, attribute cannot be set!")
                self._meta_attributes[key] = value

            def __setitem__(self, key, value):
                if self._meta_attributes is None:
                    raise t.FileModeError("File not writable, item cannot be set!")
                setattr(self._meta_attributes, key, value)

            def __contains__(self, item):
                if self._meta_attributes is None:
                    return False
                try:
                    _ = self._meta_attributes[item]
                    return True
                except KeyError:
                    return False

            def keys(self):
                return vars(self._meta_attributes).keys()

            def values(self):
                return vars(self._meta_attributes).values()

        # existing?
        try:
            meta_group = self.file.get_node('/' + self._meta_group_name)
        except t.NoSuchNodeError:
            try:
                meta_group = self.file.create_group('/', self._meta_group_name)
            except t.FileModeError:
                logger.debug("File not open for writing, not creating meta group.")
                self.meta = MetaAccess()
                return

        try:
            meta_node = self.file.get_node(meta_group, 'meta_node')
        except t.NoSuchNodeError:
            try:
                meta_node = filenode.new_node(self.file, where=meta_group, name='meta_node')
            except t.FileModeError:
                logger.debug("File not open for writing, not creating meta node.")
                self.meta = MetaAccess()
                return

        self.meta = MetaAccess(meta_node.attrs)

    def _update_classid(self, force=False):
        if '_classid' not in self.meta or force:
            try:
                self.meta._classid = self._classid
            except (AttributeError, t.FileModeError):
                pass

    def close(self, copy_tmp=True, remove_tmp=True):
        if not self.file.isopen:
            warnings.warn("File {} is already closed!".format(self.file.filename))
            return
        file_mode = self.file.mode
        self.file.close()
        if self.tmp_file_name is not None:
            if copy_tmp and self.file_name is not None and file_mode not in ('r', 'r+'):
                logger.info("Moving temporary output file to destination {}".format(self.file_name))
                shutil.copyfile(self.tmp_file_name, self.file_name)
            if remove_tmp:
                os.remove(self.tmp_file_name)

    def _init_file(self, file_name, mode):
        if file_name is None:
            self.file = create_or_open_pytables_file()
        elif type(file_name) == str:
            self.file = create_or_open_pytables_file(file_name, mode=mode)
        elif isinstance(file_name, t.file.File):
            self.file = file_name
        else:
            raise TypeError("file_name is not a recognisable type")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        report_exception = exc_type is None
        self.close(copy_tmp=report_exception)
        return report_exception


class FileGroup(FileBased):

    _classid = 'FILEGROUP'

    def __init__(self, group, file_name=None, mode='a', tmpdir=None):
        FileBased.__init__(self, file_name=file_name, mode=mode, tmpdir=tmpdir)

        try:
            group_node = self.file.get_node("/" + group)
            if not isinstance(group_node, t.group.Group):
                raise TypeError("%s is not a group, but %s" % (group, str(type(group_node))))
            self._group = group_node
        except t.NoSuchNodeError:
            self._group = self.file.create_group('/', group)


class TableObject(object):
    def __getitem__(self, key):
        try:
            return self.__getattribute__(key)
        except AttributeError:
            raise IndexError("No item " + str(key) + " in object")


class Mask(object):
    """
    Class providing Mask details.
    """
    
    def __init__(self, name=None, description='', ix=None):
        if ix is not None:
            self.ix = ix
        else:
            self.ix = 0
        self.name = name.decode() if isinstance(name, bytes) else name
        self.description = description.decode() if isinstance(description, bytes) else description

    def __repr__(self):
        return "%d. %s: %s" % (self.ix, self.name, self.description)


class Maskable(FileBased):
    """
    Class that adds masking functionality to tables.
    
    It can be desirable to hide a subset of data in 
    a table rather than deleting it outright. This is 
    especially useful in pytables tables, as they are
    not designed for frequent deletions and may 
    suffer from performance problems as a consequence.
    The process of hiding data is referred to as 
    'masking'.
    
    This class creates a pytables Table that contains
    meta-information about masks, i.e. mask IDs, mask
    names, and mask descriptions. Since it is backed
    by pytables, Mask descriptions will be saved to 
    the corresponding hdf5 file automatically. This 
    class additionally provides tools to create and 
    retrieve Mask descriptions.
    
    Note that this class only provides meta-
    information for Masks. Masking can technically 
    performed without Maskable, but then the reasons
    for masking will not be saved.
    
    The actual masking is performed using a MaskFilter
    on a MaskedTable object using the 
    MaskedTable.filter(MaskFilter) function. 
    """

    _classid = 'MASKABLE'
    
    class MaskDescription(t.IsDescription):
        ix = t.Int16Col(pos=0)
        name = t.StringCol(50, pos=1)
        description = t.StringCol(255, pos=2)

    def __init__(self, data=None, file_name=None, table_name="mask", mode='a', tmpdir=None):
        """
        Enable recording of masking in pytables-backed object.
        
        This constructor is built with considerable flexibility
        in the ways it can be called. The 'data' parameter
        serves here as a multi-purpose container that simplifies
        calling the constructor with just one argument.
        
        Args:
            data (...):         multi-purpose container, simplifies one-argument
                                constructor calls. Can stand in for a pytables 
                                Table or File, or a path.
                (None):         An existing mask description table is expected in 
                                self._mask. If none is found, the constructor will 
                                look for a pytables File in self.file and attempt 
                                to create the mask description table in there.
                (str):          An exisiting pytable File at this location will be
                                opened in append mode or one will be created.
                (tables.file.File):
                                A mask description table will be created in this
                                pytables File.
                (tables.table.Table):
                                This pytables Table will be used as mask 
                                description table (if it has the necessary fields
                                ix, name, description).
            table_name (str):   name of the mask table that is created in
                                pytables file, does not usually need to be 
                                modified
        """
        # parse potential unnamed argument
        if data is not None:
            # data is file name
            if type(data) is str or isinstance(data, t.file.File):
                if file_name is None:
                    file_name = data
                    data = None
            elif type(data) == t.table.Table:
                self._set_mask_table(data)
        else:
            if hasattr(self, '_mask'):
                self._set_mask_table(self._mask)

        FileBased.__init__(self, file_name, tmpdir=tmpdir, mode=mode)
                
        if (not hasattr(self, '_mask') or self._mask is None) and self.file is not None:
            try:
                self._mask = self.file.get_node('/' + table_name)
            except NoSuchNodeError:
                self._mask = self.file.create_table("/", table_name, Maskable.MaskDescription)
                row = self._mask.row
                row['ix'] = 0
                row['name'] = 'default'
                row['description'] = 'Default mask'
                row.append()
                self._mask.flush()
        
    def _set_mask_table(self, table):
        if type(table) == t.table.Table:
            if ('ix' not in table.colnames or
                    'name' not in table.colnames or
                    'description' not in table.colnames):
                raise ValueError("Object already has a mask table, \
                                  but it does not have all the necessary \
                                  columns (ix, name, description)")
            self._mask = table
        else:
            raise ValueError("Table is not a MaskDescription table")

    def add_mask_description(self, name, description):
        """
        Add a mask description to the _mask table and return its ID.
        

        Args:
            name (str):        name of the mask
            description (str): description of the mask
            
        Returns:
            int:    id of the mask
        """
        ix = len(self._mask)
        row = self._mask.row
        row['ix'] = ix
        row['name'] = name
        row['description'] = description
        row.append()
        self._mask.flush()

        return Mask(ix=ix, name=name, description=description)

    def masks(self):
        this = self

        class MaskIter(object):
            def __init__(self):
                self.iter = iter(this._mask)

            def __iter__(self):
                return self

            def __next__(self):
                row = next(self.iter)
                return Maskable._row_to_mask(row)

        return MaskIter()

    @staticmethod
    def _row_to_mask(row):
        return Mask(name=row['name'], ix=row['ix'], description=row['description'])

    def get_binary_mask_from_masks(self, masks):
        o = []
        for m in masks:
            if isinstance(m, string_types):
                o.append(2**self.get_mask(m).ix)
            elif isinstance(m, MaskFilter):
                o.append(2**m.mask_ix)
            elif isinstance(m, Mask):
                o.append(2**m.ix)
            elif isinstance(m, int):
                o.append(2**m)
            else:
                raise ValueError('Can only get binary mask from mask names, indexes and MaskFilter instances')
        return sum(o)

    def get_mask(self, key):
        """
        Search _mask table for key and return Mask.
        
        Args:
            key (str): search by mask name
            key (int): search by mask ID
            
        Returns:
            Mask
        """
        
        if isinstance(key, int):
            for row in self._mask.where("ix == %d" % key):
                return Maskable._row_to_mask(row)
        else:
            for row in self._mask.where("name == b'%s'" % str(key)):
                return Maskable._row_to_mask(row)
        return KeyError("Unrecognised key type")
    
    def get_masks(self, ix):
        """
        Extract mask IDs encoded in parameter and return masks.
        
        IDs are powers of 2, so a single int field in the table can hold
        multiple masks by simply adding up the IDs. Similar principle to
        UNIX chmod (although that uses base 8)
        
        Args:
            ix (int): integer that is the sum of powers of 2. Note that this value
                      is not necessarily itself a power of 2.
            
        Returns:
            list (Mask): list of Masks extracted from ix
        """
        
        key_copy = ix

        last_ix = int(self._mask[-1][0])
        
        masks = []
        while last_ix > 0:
            if key_copy - 2**last_ix >= 0:
                mask = self.get_mask(last_ix)
                if mask is not None:
                    masks.append(mask)
                key_copy -= 2**last_ix
            last_ix -= 1
        if key_copy > 0:
            masks.append(self.get_mask(0))
        
        return list(reversed(masks))

    def _mask_statistics_table(self, table, include_unmasked=True):
        masks = defaultdict(int)
        if include_unmasked:
            masks['unmasked'] = 0

        if 'mask_stats' in table.attrs:
            stats = table.attrs['mask_stats']
        else:
            stats = table.mask_stats()

        for mask_bit, count in stats.items():
            row_masks = self.get_masks(mask_bit)

            found_masks = False
            for mask in row_masks:
                found_masks = True
                masks[mask.name] += count

            if not found_masks and include_unmasked:
                masks['unmasked'] += count
        return masks

    def _mask_statistics_group(self, group, include_unmasked=True):
        masks = defaultdict(int)

        for table in group:
            for mask, count in self._mask_statistics_table(table, include_unmasked=include_unmasked).items():
                masks[mask] += count

        return masks

    def mask_statistics(self, table, include_unmasked=True):
        if isinstance(table, t.Group):
            return self._mask_statistics_group(table, include_unmasked=include_unmasked)
        elif isinstance(table, MaskedTable):
            return self._mask_statistics_table(table, include_unmasked=include_unmasked)
        else:
            raise ValueError("First arg must be PyTable Group or MaskedTable!")


class MaskedTableView(object):
    def __init__(self, masked_table, it=None, excluded_masks=0):
        self.masked_table = masked_table
        self.iter = it if it else self.masked_table._iter_visible_and_masked()
        self.excluded_mask_ix = excluded_masks

    def __iter__(self):
        return self

    def __next__(self):
        row = next(self.iter)
        # bit-shift magic! Go @alexis!
        # a is a subset of b if and only if a | b == b.
        # If this condition is satisfied for each byte, return TRUE. Otherwise return FALSE
        while row[self.masked_table._mask_field] | self.excluded_mask_ix != self.excluded_mask_ix:
            row = next(self.iter)
        return row


class MaskedTable(t.Table):
    """
    Wrapper that adds masking functionality to a pytables table. 
    
            
    MaskedTable is simply a wrapper around a
    pytables Table with certain columns. It
    adds masking functionality to table rows.
        
    It can be desirable to hide a subset of data in 
    a table rather than deleting it outright. This is 
    especially useful in pytables tables, as they are
    not designed for frequent deletions and may 
    suffer from performance problems as a consequence.
    
    This class extends the functionality of classes
    that work on the basis of a pytables table. It
    allows the masking of rows and provides table iterators
    and len() functions that ignore masked rows, and 
    therefore give the illusion of deleted rows.
    Masked rows, however, can be inspected separately
    at any point.
    
    The minimum requirement for a table to be masked
    is the inclusion of a 'mask' column. For full
    iterator and len() functionality, the table also
    needs to include a 'mask_ix' column. Checks for
    this are performed when specifying the table to
    be masked.
    """
    
    _c_classid = 'MASKEDTABLE'
    
    def __init__(self, parentnode, name, description=None,
                 title="", filters=None, expectedrows=None,
                 chunkshape=None, byteorder=None, _log=False,
                 mask_field='_mask', mask_index_field='_mask_ix',
                 ignore_reserved_fields=False):
        """
        Pytables Table extension to provide masking functionality.
        """
        
        # set instance variables
        self._queued_filters = []
        self._mask_field = mask_field
        self._mask_index_field = mask_index_field

        if description is not None:
            # try converting description to dict
            try:
                description = description.columns
            except AttributeError:
                pass
            
            # fill in fields required for masking
            if isinstance(description, dict):
                masked_description = description.copy()
            else:
                raise ValueError("Unrecognised description type (%s)" % str(type(description)))
            
            # check that reserved keys are not used
            if mask_field in masked_description:
                if not ignore_reserved_fields:
                    raise ValueError("{} field is reserved in MaskedTable!".format(mask_field))
            else:
                masked_description[mask_field] = t.Int32Col()

            if mask_index_field in masked_description:
                if not ignore_reserved_fields:
                    raise ValueError("{} field is reserved in MaskedTable!".format(mask_index_field))
            else:
                masked_description[mask_index_field] = t.Int64Col()
        else:
            masked_description = None
                
        t.Table.__init__(self, parentnode, name,
                         description=masked_description, title=title,
                         filters=filters,
                         expectedrows=expectedrows,
                         _log=_log,
                         chunkshape=chunkshape,
                         byteorder=byteorder)
        
        self.enable_mask_index()

    def disable_mask_index(self):
        mask_ix_col = getattr(self.cols, self._mask_index_field)
        mask_ix_col.remove_index()

    def enable_mask_index(self):
        mask_ix_col = getattr(self.cols, self._mask_index_field)
        create_col_index(mask_ix_col)

    def flush(self, update_index=False, log_progress=True):
        """
        Flush buffered rows.
        
        Also updates the mask index, if requested.
        """
        self._flush(update_index, log_progress=log_progress)
    
    def _flush(self, update_index=False, log_progress=True):
        # commit any previous changes
        t.Table.flush(self)

        if update_index:
            self._update_ix(log_progress=log_progress)
            # force flush of index if
            # autoindex is disabled
            if not self.autoindex:
                self.flush_rows_to_index()
            # commit index changes
            t.Table.flush(self)

    def iterrows(self, start=None, stop=None, step=None, excluded_masks=0):
        it = t.Table.iterrows(self, start, stop, step)
        return MaskedTableView(self, it, excluded_masks=excluded_masks)

    def itersorted(self, sortby, checkCSI=False,
                   start=None, stop=None, step=None, excluded_masks=0):
        it = t.Table.itersorted(self, sortby, checkCSI=checkCSI,
                                start=start, stop=stop, step=step)
        return MaskedTableView(self, it, excluded_masks=excluded_masks)

    def _iter_visible_and_masked(self):
        """
        Return an iterator over all rows, including masked ones.
        """
        return t.Table.iterrows(self)

    def masked_rows(self):
        """
        Return an iterator over masked rows.
        """
        this = self
        it = self._iter_visible_and_masked()

        class MaskedRows(MaskedTableView):
            def __init__(self, masked_table, it_):
                super(MaskedRows, self).__init__(masked_table, it_)

            def __next__(self):
                row = next(self.iter)
                while row[this._mask_field] == 0:
                    row = next(self.iter)
                return row

            def __getitem__(self, key):
                if isinstance(key, int):
                    if key >= 0:
                        key = -1*key - 1
                        res = [x.fetch_all_fields() for x in
                               super(MaskedTable, this).where("%s == %d" % (this._mask_index_field, key))]
                        if len(res) == 1:
                            return res[0]
                        if len(res) == 0:
                            raise IndexError("Index %d out of bounds" % key)
                        raise RuntimeError("Duplicate row for key %d" % key)
                    else:
                        l = this._visible_len()
                        return self[l+key]
                else:
                    raise KeyError('Cannot retrieve row with key ' + str(key))

        return MaskedRows(self, it)

    def __getitem__(self, key):
        return self._get_visible_item(key)
    
    def _get_visible_item(self, key):
        if type(key) == slice:
            res = []
            # set sensible defaults
            start = key.start
            if start is None:
                start = 0
            stop = key.stop
            if stop is None:
                stop = len(self)
            step = key.step
            if step is None:
                step = 1
                  
            for i in range(start, stop, step):
                res.append(self._get_visible_item(i))
            return res
        else:
            try:
                # this could be a numpy int
                try:
                    key = key.item()
                except AttributeError:
                    pass
                
                key = int(key)
                
                if key >= 0:
                    res = [x.fetch_all_fields() for x in self.where("%s == %d" % (self._mask_index_field, key))]
                    if len(res) == 1:
                        return res[0]
                    if len(res) == 0:
                        raise IndexError("Index %d out of bounds" % key)
                    raise RuntimeError("Duplicate row for key %d" % key)
                else:
                    l = self._visible_len()
                    if l+key >= 0:
                        return self._get_visible_item(l+key)
                    raise KeyError('Cannot retrieve row with key %s' % str(key))
            except ValueError as e:
                raise KeyError('Cannot retrieve row with key %s (%s)' % (str(key), str(e)))
    
    def _original_getitem(self, key):
        return t.Table.__getitem__(self, key)
    
    def __len__(self):
        """
        Return the 'perceived' length of the masked table.
          
        If the table has masked rows, these will not be counted.
        """
          
        return self._visible_len()
    
    def _visible_len(self):
        if 'masked_length' not in self.attrs or self.attrs['masked_length'] == -1:
            return sum(1 for _ in iter(self.where("%s >= 0" % self._mask_index_field)))
        return int(self.attrs['masked_length'])
    
    def _original_len(self):
        return t.Table.__len__(self)
    
    # new index update method
    def _update_ix(self, log_progress=not config.hide_progressbars):
        """
        Update the row indexes of the Table.
        
        Should be run after masking. Will assign auto-
        incrementing integers (from 0) to the mask index
        field of each row in the table if it is not 
        masked, -1 otherwise.
        """

        if log_progress:
            logger.debug("Updating mask indices")

        stats = defaultdict(int)

        l = self._original_len()
        with RareUpdateProgressBar(max_value=l, silent=not log_progress) as pb:
            ix = 0
            masked_ix = -1
            for i, row in enumerate(self._iter_visible_and_masked()):
                m = row[self._mask_field]
                stats[m] += 1

                if m > 0:
                    row[self._mask_index_field] = masked_ix
                    masked_ix -= 1
                else:
                    row[self._mask_index_field] = ix
                    ix += 1
                row.update()
                pb.update(i)
            self.attrs['masked_length'] = ix
        try:
            self.attrs['mask_stats'] = stats
        except t.FileModeError:
            pass

    def mask_stats(self):
        stats = defaultdict(int)
        for row in self._iter_visible_and_masked():
            stats[row[self._mask_field]] += 1

        try:
            self.attrs['mask_stats'] = stats
        except t.FileModeError:
            pass

        return stats
    
    def _get_masks(self, binary_mask):
        def bits(n):
            while n:
                b = n & (~n+1)
                yield b
                n ^= b
        return list(bits(binary_mask))

    def _row_masks(self, row):
        return self._get_masks(row[self._mask_field])

    def _has_mask(self, row, mask):
        return mask in self._row_masks(row)

    def filter(self, mask_filter, _logging=not config.hide_progressbars):
        """
        Run a MaskFilter on this table.
        
        This functions calls the MaskFilter.valid function on
        every row and masks them if the function returns False.
        After running the filter, the table index is updated
        to match only unmasked rows.

        :param mask_filter: A :class:`~MaskFilter` object
        :param _logging: Print progress to stderr
        """

        total = 0
        ix = 0
        mask_ix = -1

        # statistics
        stats = defaultdict(int)
        with RareUpdateProgressBar(max_value=self._original_len(), silent=not _logging) as pb:
            for i, row in enumerate(self._iter_visible_and_masked()):
                total += 1

                if not mask_filter.valid(row):
                    row[self._mask_field] += 2**mask_filter.mask_ix

                stats[row[self._mask_field]] += 1

                # update index
                if row[self._mask_field] > 0:
                    row[self._mask_index_field] = mask_ix
                    mask_ix -= 1
                else:
                    row[self._mask_index_field] = ix
                    ix += 1
                row.update()
                pb.update(i)
        self.attrs['masked_length'] = ix

        if _logging:
            logger.info("Total: %d. Filtered: %d" % (total, -1*(mask_ix-1)))

        try:
            self.attrs['mask_stats'] = stats
        except t.FileModeError:
            pass

        self.flush(update_index=False)

        return stats

    def queue_filter(self, filter_definition):
        """
        Add a MaskFilter to filter queue.
        
        Queued filters can be run at a later time using
        the run_queued_filters function.
        """
        self._queued_filters.append(filter_definition)
        
    def run_queued_filters(self, _logging=not config.hide_progressbars):
        """
        Run queued MaskFilters.
        
        MaskFilters can be queued using the
        queue_filter function.

        :param _logging: If True, prints log to stderr
        """
        ix = 0
        mask_ix = -1
        total = 0

        stats = defaultdict(int)
        with RareUpdateProgressBar(max_value=self._original_len(), silent=not _logging) as pb:
            for i, row in enumerate(self._iter_visible_and_masked()):
                total += 1
                for f in self._queued_filters:
                    if not f.valid(row):
                        row[self._mask_field] += 2**f.mask_ix

                stats[row[self._mask_field]] += 1

                # update index
                if row[self._mask_field] > 0:
                    row[self._mask_index_field] = mask_ix
                    mask_ix -= 1
                else:
                    row[self._mask_index_field] = ix
                    ix += 1
                row.update()
                pb.update(i)

        self.attrs['masked_length'] = ix

        if _logging:
            logger.info("Total: %d. Filtered: %d" % (total, -1*(mask_ix-1)))

        try:
            self.attrs['mask_stats'] = stats
        except t.FileModeError:
            pass

        self.flush(update_index=False)
        return stats

    def where(self, condition, condvars=None,
              start=None, stop=None, step=None):
        condition = "(" + condition + ") & (%s >= 0)" % self._mask_index_field
        return super(MaskedTable, self).where(condition, condvars=None,
                                              start=None, stop=None, step=None)


class MaskFilter(with_metaclass(ABCMeta, object)):
    """
    Abstract class that defines a filter for MaskedTable.
    """

    def __init__(self, mask=None, mask_ix=0, mask_name='default', mask_description="Default mask."):
        """
        Create a MaskFilter.
        
        Sets values for mask index, name, and description
        to be used by MaskedTable.filter function.
        
        Args:
            mask (Mask, int):    if Mask, the attributes mask_ix,
                                 mask_name, and mask_description
                                 will be copied here.
                                 if int, mask_name and mask_description
                                 will be set to defaults, while
                                 mask_ix will be set to this value.
            mask_ix (int):       Index of the mask to be applied.
                                 Defaults to 0
            mask_name (str):     Name of the mask to be applied.
                                 Defaults to 'default'
            mask_description (str):
                                 Description of the mask to be applied.
                                 Defaults to 'Default mask'
        """
        
        if mask is not None:
            if isinstance(mask, Mask):
                self.mask_ix = mask.ix
                self.mask_name = mask.name
                self.mask_description = mask.description
            else:
                self.mask_ix = mask
                self.mask_name = mask_name
                self.mask_description = mask_description
        else:
            self.mask_ix = mask_ix
            self.mask_name = mask_name
            self.mask_description = mask_description
            
        if not type(self.mask_ix) == int:
            raise ValueError("mask_ix must be an integer!")
    
    @abstractmethod
    def valid(self, row):
        """
        Test if a row is valid according to this filter.
        
        Args:
            row (tables.tableextension.Row):
                A pytables Table Row
        
        Returns:
            bool: True if row is valid, False otherwise
        """
        pass
